#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

""":term:`MKID`-array detector modeling class."""

import typing as t
from pathlib import Path

import h5py as h5
import numpy as np

from pyxel import __version__
from pyxel.data_structure import Phase
from pyxel.detectors import Detector
from pyxel.util.memory import memory_usage_details

if t.TYPE_CHECKING:
    from pyxel.detectors import Environment, MKIDCharacteristics, MKIDGeometry


class MKID(Detector):
    """:term:`MKID`-based detector class containing all detector attributes and data."""

    def __init__(
        self,
        geometry: "MKIDGeometry",
        environment: "Environment",
        characteristics: "MKIDCharacteristics",
    ):
        self._geometry = geometry  # type: MKIDGeometry
        self._characteristics = characteristics  # type: MKIDCharacteristics
        self._phase = None  # type: t.Optional[Phase]

        super().__init__(environment=environment)
        self.reset()

    def reset(self) -> None:
        """TBW."""
        super().reset()
        self._phase = Phase(geo=self.geometry)

    def empty(self, empty_all: bool = True) -> None:
        """Empty the data in the detector.

        Returns
        -------
        None
        """
        super().empty(empty_all)

        if empty_all and self._phase:
            self.phase.array *= 0

    @property
    def geometry(self) -> "MKIDGeometry":
        """TBW."""
        return self._geometry

    @property
    def characteristics(self) -> "MKIDCharacteristics":
        """TBW."""
        return self._characteristics

    @property
    def phase(self) -> Phase:
        """TBW."""
        if not self._phase:
            raise RuntimeError("'phase' not initialized.")

        return self._phase

    def memory_usage(
        self, print_result: bool = True, human_readable: bool = True
    ) -> dict:
        """TBW.

        Returns
        -------
        dict
            Dictionary of attribute memory usage
        """
        attributes = [
            "_photon",
            "_charge",
            "_pixel",
            "_signal",
            "_image",
            "_phase",
            "material",
            "environment",
            "_geometry",
            "_characteristics",
        ]

        return memory_usage_details(
            self, attributes, print_result=print_result, human_readable=human_readable
        )

    # TODO: Move this to another place. See #241
    def to_hdf5(self, filename: t.Union[str, Path]) -> None:
        """Convert the detector to a HDF5 object."""
        with h5.File(filename, "w") as h5file:
            h5file.attrs["type"] = self.__class__.__name__
            h5file.attrs["pyxel-version"] = str(__version__)
            detector_grp = h5file.create_group("detector")
            for array, name in zip(
                [
                    self.signal.array,
                    self.image.array,
                    self.photon.array,
                    self.pixel.array,
                    self.phase.array,
                    self.charge.frame,
                ],
                ["Signal", "Image", "Photon", "Pixel", "Phase", "Charge"],
            ):
                dataset = detector_grp.create_dataset(name, shape=np.shape(array))
                dataset[:] = array



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Poppy model."""

import logging
import typing as t
from dataclasses import dataclass

import numpy as np
from astropy.convolution import convolve_fft
from astropy.io import fits

from pyxel.detectors import Detector

try:
    import poppy as op

    WITH_POPPY: bool = True
except ImportError:
    WITH_POPPY = False


@dataclass
class CircularAperture:
    """Parameters for an ideal circular pupil aperture.

    Parameters
    ----------
    radius : float
        Radius of the pupil, in meters.
    """

    radius: float


@dataclass
class ThinLens:
    """Parameters for an idealized thin lens.

    Parameters
    ----------
    nwaves : float
        The number of waves of defocus, peak to valley.
    radius : float
        Pupil radius, in meters, over which the Zernike defocus term should be computed
        such that rho = 1 at r = `radius`.
    """

    nwaves: float
    radius: float


@dataclass
class SquareAperture:
    """Parameters for an ideal square pupil aperture.

    Parameters
    ----------
    size : float
        side length of the square, in meters.
    """

    size: float


@dataclass
class RectangleAperture:
    """Parameters for an ideal rectangular pupil aperture.

    Parameters
    ----------
    width : float
        width of the rectangle, in meters.
    height : float
        height of the rectangle, in meters.
    """

    width: float
    height: float


@dataclass
class HexagonAperture:
    """Parameters for an ideal hexagonal pupil aperture.

    Parameters
    ----------
    side : float
        side length (and/or radius) of hexagon, in meters.
    """

    side: float


@dataclass
class MultiHexagonalAperture:
    """Parameters for an hexagonaly segmented aperture.

    Parameters
    ----------
    side : float
        side length (and/or radius) of hexagon, in meters.
    rings : integer
        The number of rings of hexagons to include, not counting the central segment
        (i.e. 2 for a JWST-like aperture, 3 for a Keck-like aperture, and so on)
    gap : float
        Gap between adjacent segments, in meters.
    """

    side: float
    rings: int
    gap: float


@dataclass
class SecondaryObscuration:
    """Parameters to define the central obscuration of an on-axis telescope.

    The parameters include secondary mirror and supports.

    Parameters
    ----------
    secondary_radius : float
        Radius of the circular secondary obscuration, in meters.
    n_supports : int
        Number of secondary mirror supports ("spiders"). These will be
        spaced equally around a circle.
    support_width : float
        Width of each support, in meters.
    """

    secondary_radius: float
    n_supports: int
    support_width: float


@dataclass
class ZernikeWFE:
    """Parameters to define an optical element in terms of its Zernike components.

    Parameters
    ----------
    radius : float
        Pupil radius, in meters, over which the Zernike terms should be
        computed such that rho = 1 at r = `radius`.
    coefficients : iterable of floats
        Specifies the coefficients for the Zernike terms, ordered
        according to the convention of Noll et al. JOSA 1976. The
        coefficient is in meters of optical path difference (not waves).
    aperture_stop : float
    """

    radius: float
    coefficients: t.Sequence[float]
    aperture_stop: float


@dataclass
class SineWaveWFE:
    """Parameters to define a single sine wave ripple across the optic.

    Parameters
    ----------
    spatialfreq : float
    amplitude : float
    rotation : float
    """

    spatialfreq: float
    amplitude: float
    rotation: float


# Define a type alias
OpticalParameter = t.Union[
    CircularAperture,
    ThinLens,
    SquareAperture,
    RectangleAperture,
    HexagonAperture,
    MultiHexagonalAperture,
    SecondaryObscuration,
    ZernikeWFE,
    SineWaveWFE,
]


def create_optical_parameter(dct: t.Mapping) -> OpticalParameter:
    """Create a new ``OpticalParameter`` based on a dictionary.

    Parameters
    ----------
    dct : dict
        Dictionary to convert

    Returns
    -------
    OpticalParameter
        New parameters.
    """
    if dct["item"] == "CircularAperture":
        return CircularAperture(radius=dct["radius"])

    elif dct["item"] == "ThinLens":
        return ThinLens(nwaves=dct["nwaves"], radius=dct["radius"])

    elif dct["item"] == "SquareAperture":
        return SquareAperture(size=dct["size"])

    elif dct["item"] == "RectangularAperture":
        return RectangleAperture(width=dct["width"], height=dct["height"])

    elif dct["item"] == "HexagonAperture":
        return HexagonAperture(side=dct["side"])

    elif dct["item"] == "MultiHexagonalAperture":
        return MultiHexagonalAperture(
            side=dct["side"],
            rings=dct["rings"],
            gap=dct["gap"],
        )  # cm

    elif dct["item"] == "SecondaryObscuration":
        return SecondaryObscuration(
            secondary_radius=dct["secondary_radius"],
            n_supports=dct["n_supports"],
            support_width=dct["support_width"],
        )  # cm

    elif dct["item"] == "ZernikeWFE":
        return ZernikeWFE(
            radius=dct["radius"],
            coefficients=dct["coefficients"],  # list of floats
            aperture_stop=dct["aperture_stop"],
        )  # bool

    elif dct["item"] == "SineWaveWFE":
        return SineWaveWFE(
            spatialfreq=dct["spatialfreq"],  # 1/m
            amplitude=dct["amplitude"],  # um
            rotation=dct["rotation"],
        )
    else:
        raise NotImplementedError


def create_optical_parameters(
    optical_system: t.Sequence[t.Mapping],
) -> t.Sequence[OpticalParameter]:
    """Create a list of ``OpticalParameters``.

    Parameters
    ----------
    optical_system : ``list`` of ``dict``
        List to convert.

    Returns
    -------
    ``list`` of ``OpticalParameter``
        A new list of parameters.
    """
    return [create_optical_parameter(dct) for dct in optical_system]


def create_optical_item(
    param: OpticalParameter,
    wavelength: float,
) -> "op.OpticalElement":
    """Create a new poppy ``OpticalElement``.

    Parameters
    ----------
    param : ``OpticalParameter``
        Pyxel Optical parameters to create a poppy ``OpticalElement``.
    wavelength : float

    Returns
    -------
    ``OpticalElement``
        A new poppy ``OpticalElement``.
    """
    if isinstance(param, CircularAperture):
        return op.CircularAperture(radius=param.radius)

    elif isinstance(param, ThinLens):
        return op.ThinLens(
            nwaves=param.nwaves,
            reference_wavelength=wavelength,
            radius=param.radius,
        )

    elif isinstance(param, SquareAperture):
        return op.SquareAperture(size=param.size)

    elif isinstance(param, RectangleAperture):
        return op.RectangleAperture(width=param.width, height=param.height)

    elif isinstance(param, HexagonAperture):
        return op.HexagonAperture(side=param.side)

    elif isinstance(param, MultiHexagonalAperture):
        return op.MultiHexagonAperture(
            side=param.side, rings=param.rings, gap=param.gap
        )

    elif isinstance(param, SecondaryObscuration):
        return op.SecondaryObscuration(
            secondary_radius=param.secondary_radius,
            n_supports=param.n_supports,
            support_width=param.support_width,
        )

    elif isinstance(param, ZernikeWFE):
        return op.ZernikeWFE(
            radius=param.radius,
            coefficients=param.coefficients,
            aperture_stop=param.aperture_stop,
        )

    elif isinstance(param, SineWaveWFE):
        return op.SineWaveWFE(
            spatialfreq=param.spatialfreq,
            amplitude=param.amplitude,
            rotation=param.rotation,
        )
    else:
        raise NotImplementedError


def calc_psf(
    wavelength: float,
    fov_arcsec: float,
    pixelscale: float,
    optical_parameters: t.Sequence[OpticalParameter],
) -> t.Tuple[t.Sequence[fits.hdu.image.PrimaryHDU], t.Sequence["op.Wavefront"]]:
    """Calculate the point spread function for the given optical system.

    Parameters
    ----------
    wavelength : float
        Wavelength of incoming light in meters.
    fov_arcsec : float, optional
        Field Of View on detector plane in arcsec.
    pixelscale : float
        Pixel scale on detector plane (arcsec/pixel).
        Defines sampling resolution of :term:`PSF`.
    optical_parameters : list of OpticalParameter
        List of optical parameters before detector with their specific arguments.

    Returns
    -------
    Sequence of FITS and sequence of Wavefront
        Tuple of lists containing the psf and intermediate wavefronts.
    """
    if not WITH_POPPY:
        raise ImportError(
            "Missing optional package 'poppy'.\n"
            "Please install it with 'pip install pyxel-sim[model]' "
            "or 'pip install pyxel-sim[all]'"
        )

    # Create the optical element(s)
    osys = op.OpticalSystem(npix=1000)  # default: 1024

    for param in optical_parameters:  # type: OpticalParameter
        element = create_optical_item(
            param=param,
            wavelength=wavelength,
        )  # type: op.OpticalElement

        osys.add_pupil(element)

    osys.add_detector(
        pixelscale=pixelscale,
        fov_arcsec=fov_arcsec,
    )

    # Calculate a monochromatic PSF
    output_fits, wavefronts = osys.calc_psf(
        wavelength=wavelength,
        return_intermediates=True,
        normalize="last",
    )  # type: t.Sequence[fits.hdu.image.PrimaryHDU], t.Sequence[op.Wavefront]

    return output_fits, wavefronts


def apply_convolution(data_2d: np.ndarray, kernel_2d: np.ndarray) -> np.ndarray:
    """Convolve an array.

    Parameters
    ----------
    data_2d : ndarray
        2D Array to be convolved with kernel_2d.
    kernel_2d : ndarray
        The convolution kernel.

    Returns
    -------
    ndarray
        A convolved array.
    """
    mean = np.mean(data_2d)

    array_2d = convolve_fft(
        data_2d,
        kernel=kernel_2d,
        boundary="fill",
        fill_value=mean,
    )

    return array_2d


def optical_psf(
    detector: Detector,
    wavelength: float,
    fov_arcsec: float,
    pixelscale: float,
    optical_system: t.Sequence[t.Mapping],
) -> None:
    """Model function for poppy optics model: convolve photon array with psf.

    Parameters
    ----------
    detector : Detector
        Pyxel Detector object.
    wavelength : float
        Wavelength of incoming light in meters.
    fov_arcsec : float
        Field Of View on detector plane in arcsec.
    pixelscale : float
        Pixel scale on detector plane (arcsec/pixel).
        Defines sampling resolution of :term:`PSF`.
    optical_system : list of dict
        List of optical elements before detector with their specific arguments.
    """
    logging.getLogger("poppy").setLevel(
        logging.WARNING
    )  # TODO: Fix this. See issue #81

    # Validation and Conversion stage
    # These steps will be probably moved into the YAML engine
    if wavelength < 0.0 or fov_arcsec < 0.0 or pixelscale < 0.0:
        raise ValueError(
            "Expecting strictly positive value for 'wavelength', "
            "'fov_arcsec' and 'pixelscale'."
        )

    # Convert 'optical_system' to 'optical_parameters'
    optical_parameters = [
        create_optical_parameter(dct) for dct in optical_system
    ]  # type: t.Sequence[OpticalParameter]

    # Processing
    # Get a Point Spread Function
    images, wavefronts = calc_psf(
        wavelength=wavelength,
        fov_arcsec=fov_arcsec,
        pixelscale=pixelscale,
        optical_parameters=optical_parameters,
    )

    # Extract 'first_image'
    first_image, *other_images = images

    # Convolution
    new_array_2d = apply_convolution(
        data_2d=detector.photon.array,
        kernel_2d=first_image.data,
    )  # type: np.ndarray

    detector.photon.array = new_array_2d



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""TBW."""

import typing as t
from dataclasses import dataclass

import numpy as np

from pyxel.detectors import CCD

try:
    import arcticpy as ac

    WITH_ARTICPY = True
except ImportError:
    # No 'arcticpy' library
    WITH_ARTICPY = False


@dataclass
class Trap:
    """Define a trap.

    Parameters
    ----------
    density : float
        The density of the trap species in a pixel.
    release_timescale : float
        The release timescale of the trap.
    """

    density: float
    release_timescale: float


def compute_arctic_add(
    image_2d: np.ndarray,
    full_well_depth: float,
    well_fill_power: float,
    parallel_traps: t.Sequence[Trap],
    parallel_express: int,
) -> np.ndarray:
    """Create a new image with :term:`CTI` trails.

    Parameters
    ----------
    image_2d : ndarray
        2D image to process.
    full_well_depth : float
    well_fill_power : float
    parallel_traps : sequence of Traps
        List of trap to process.
    parallel_express : int

    Returns
    -------
    ndarray
        2D array with :term:`CTI` trails.
    """
    ccd = ac.CCD(
        phases=[
            ac.CCDPhase(
                full_well_depth=full_well_depth,
                well_fill_power=well_fill_power,
            )
        ]
    )

    roe = ac.ROE()

    # Create the trap(s)
    traps = [
        ac.TrapInstantCapture(
            density=trap.density,
            release_timescale=trap.release_timescale,
        )
        for trap in parallel_traps
    ]  # type: t.Sequence[ac.TrapInstantCapture]

    image_cti_added_2d = ac.add_cti(
        image=image_2d,
        parallel_traps=traps,
        parallel_ccd=ccd,
        parallel_roe=roe,
        parallel_express=parallel_express,
        verbosity=0,
    )

    return image_cti_added_2d


def arctic_add(
    detector: CCD,
    well_fill_power: float,
    trap_densities: t.Sequence[float],
    trap_release_timescales: t.Sequence[float],
    express: int = 0,
) -> None:
    """Add :term:`CTI` trails to an image by trapping, releasing and moving electrons.

    Parameters
    ----------
    detector : CCD
        Pyxel :term:`CCD` Detector object.
    well_fill_power : float
    trap_densities : sequence of float
        A 1D arrays of all trap species densities for serial clocking.
    trap_release_timescales : sequence of float
        A 1D arrays of all trap release timescales for serial clocking.
    express : int
        As described in more detail in :cite:p:`2014:massey` section 2.1.5, the effects
        of each individual pixel-to-pixel transfer can be very similar, so multiple
        transfers can be computed at once for efficiency.
        The ``express`` input sets the number of times the transfers are calculated.

            * ``express = 1`` is the fastest and least accurate.
            * ``express = 2`` means the transfers are re-computed half-way through readout.
            * ``express = N`` where ``N`` is the total number of pixels.

        Default ``express = 0`` is a convenient input for automatic ``express = N``.

    Notes
    -----
    The external library `arcticpy <https://github.com/jkeger/arcticpy>`_ is used to add
    the :term:`CTI` trails.
    """
    # Validation
    if len(trap_densities) != len(trap_release_timescales):
        raise ValueError(
            "Expecting same number of 'trap_densities' and 'trap_release_timescales'"
        )

    if len(trap_densities) == 0:
        raise ValueError("Expecting at least one 'trap_density'.")

    # Conversion - Create a list of `Trap`
    traps = [
        Trap(density=density, release_timescale=release_timescale)
        for density, release_timescale in zip(trap_densities, trap_release_timescales)
    ]  # type: t.Sequence[Trap]

    if not WITH_ARTICPY:
        raise RuntimeError(
            "ArCTIC python wrapper is not installed ! "
            "See https://github.com/jkeger/arctic"
        )

    image_cti_added_2d = compute_arctic_add(
        image_2d=np.asarray(detector.pixel.array, dtype=float),
        full_well_depth=detector.characteristics.full_well_capacity,
        well_fill_power=well_fill_power,
        parallel_traps=traps,
        parallel_express=express,
    )

    detector.pixel.array = image_cti_added_2d


def compute_arctic_remove(
    image_2d: np.ndarray,
    full_well_depth: float,
    well_fill_power: float,
    parallel_traps: t.Sequence[Trap],
    parallel_express: int,
    num_iterations: int,
) -> np.ndarray:
    """Create a new image with removed :term:`CTI` trails.

    Parameters
    ----------
    image_2d : ndarray
        2D image to process.
    full_well_depth : float
    well_fill_power : float
    parallel_traps : sequence of Traps
        List of trap to process.
    parallel_express : int
    num_iterations : int

    Returns
    -------
    ndarray
        2D array without :term:`CTI` trails.
    """
    ccd = ac.CCD(well_fill_power=well_fill_power, full_well_depth=full_well_depth)
    roe = ac.ROE()

    # Build the traps
    traps = []  # type: t.List[ac.Trap]
    for trap_info in parallel_traps:  # type: Trap
        trap = ac.Trap(
            density=trap_info.density, release_timescale=trap_info.release_timescale
        )
        traps.append(trap)

    # Remove CTI
    image_2d_cti_removed = ac.remove_cti(
        image=image_2d,
        iterations=num_iterations,
        parallel_traps=traps,
        parallel_ccd=ccd,
        parallel_roe=roe,
        parallel_express=parallel_express,
    )

    return image_2d_cti_removed


def arctic_remove(
    detector: CCD,
    well_fill_power: float,
    instant_traps: t.Sequence[t.Mapping[str, float]],
    num_iterations: int,
    express: int = 0,
) -> None:
    """Remove :term:`CTI` trails from an image by first modelling the addition of :term:`CTI`.

    Parameters
    ----------
    detector : CCD
        Pyxel :term:`CCD` Detector object.
    well_fill_power : float
    instant_traps : sequence of mapping
        A sequence of all trap species for parallel clocking.
    num_iterations : int
        Number of iterations for the forward modelling.
        More iterations provide higher accuracy at the cost of longer runtime.
        In practice, just 1 to 3 iterations are usually sufficient.
    express : int
        As described in more detail in :cite:p:`2014:massey` section 2.1.5, the effects
        of each individual pixel-to-pixel transfer can be very similar, so multiple
        transfers can be computed at once for efficiency.
        The ``express`` input sets the number of times the transfers are calculated.

            * ``express = 1`` is the fastest and least accurate.
            * ``express = 2`` means the transfers are re-computed half-way through readout.
            * ``express = N`` where ``N`` is the total number of pixels.

        Default ``express = 0`` is a convenient input for automatic ``express = N``.
    """
    # Validation
    if num_iterations > 0:
        raise ValueError("Number of iterations must be > 1.")

    # Conversion
    traps = []  # type: t.List[Trap]
    for item in instant_traps:
        if "density" not in item:
            raise KeyError("Missing key 'density' in parameter 'instant_traps'.")
        if "release_timescale" not in item:
            raise KeyError(
                "Missing key 'release_timescale' in parameter 'instant_traps'."
            )

        trap = Trap(
            density=float(item["density"]),
            release_timescale=float(item["release_timescale"]),
        )

        traps.append(trap)

    if not WITH_ARTICPY:
        raise RuntimeError(
            "ArCTIC python wrapper is not installed ! "
            "See https://github.com/jkeger/arctic"
        )

    image_2d_cti_removed = compute_arctic_remove(
        image_2d=np.asarray(detector.pixel.array, dtype=float),
        full_well_depth=detector.characteristics.full_well_capacity,
        well_fill_power=well_fill_power,
        parallel_traps=traps,
        parallel_express=express,
        num_iterations=num_iterations,
    )

    detector.pixel.array = image_2d_cti_removed



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""TBW."""

# flake8: noqa
from .array import Array
from .photon import Photon
from .pixel import Pixel
from .signal import Signal
from .charge import Charge
from .image import Image
from .phase import Phase
from .persistence import Persistence, SimplePersistence



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Pyxel photon generator models: photon shot noise."""
import typing as t

import numpy as np
from typing_extensions import Literal

from pyxel.detectors import Detector
from pyxel.util import temporary_random_state


def compute_poisson_noise(array: np.ndarray) -> np.ndarray:
    """Compute Poisson noise using the input array.

    Parameters
    ----------
    array: ndarray
        Input array.

    Returns
    -------
    output: ndarray
        Output array.
    """
    output = np.random.poisson(lam=array).astype(array.dtype)
    return output


def compute_gaussian_noise(array: np.ndarray) -> np.ndarray:
    """Compute Gaussian noise using the input array. Standard deviation is square root of the values.

    Parameters
    ----------
    array: ndarray
        Input array.

    Returns
    -------
    output: ndarray
        Output array.
    """
    output = np.random.normal(loc=array, scale=np.sqrt(array))
    return output


def compute_noise(array: np.ndarray, type: str = "poisson") -> np.ndarray:
    """Compute shot noise for an input array. It can be either Poisson noise or Gaussian.

    Parameters
    ----------
    array: ndarray
        Input array.
    type: str, optional
        Choose either 'poisson' or 'normal'. Default is Poisson noise.

    Returns
    -------
    output: ndarray
        Output array.
    """
    if type == "poisson":
        output = compute_poisson_noise(array)
        return output
    elif type == "normal":
        output = compute_gaussian_noise(array)
        return output
    else:
        raise ValueError("Invalid noise type!")


@temporary_random_state
def shot_noise(
    detector: Detector,
    type: Literal["poisson", "normal"] = "poisson",
    seed: t.Optional[int] = None,
) -> None:
    """Add shot noise to the flux of photon per pixel. It can be either Poisson noise or Gaussian.

    Parameters
    ----------
    detector: Detector
        Pyxel Detector object.
    type: str, optional
        Choose either 'poisson' or 'normal'. Default is Poisson noise.
    seed: int, optional
        Random seed.
    """
    noise_array = compute_noise(array=detector.photon.array, type=type)
    detector.photon.array = noise_array



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

# TODO: Create a simple kernel convolution diffusion model


"""Pyxel charge diffusion and collection model extracted from TARS."""
# import logging
# from math import sqrt, log
# import numpy as np
# from astropy import units as u
# from scipy.special import erf
# from astropy.units import cds
# from pyxel.detectors.ccd import CCD
# cds.enable()


# # @validators.validate
# # @config.argument(name='', label='', units='', validate=)
# def diffusion(detector: CCD):
#     """Diffusing electrons with some random charge transport."""
#     logger = logging.getLogger('pyxel')
#     logger.info('')
#
#     # diff = Diffusion(detector)
#     # collected_charge_list = []  # type: list
#     # cluster_generator = [cluster for cluster in detector.charge_list]
#     # for cluster in cluster_generator:
#     # # Modifying the positions and shapes of charge clusters in list of Charge class instances
#     # sigma_hiraga = diff.hiraga_diffusion_model(cluster)
#     # sigma_janesick = diff.janesick_diffusion_model(cluster)
#     # sigma = 1.0     # temporarily
#     # collected_charge = diff.gaussian_pixel_separation(cluster, sigma, sigma)
#     # collected_charge_list += collected_charge
#     # # Overwrite the list of charge clusters in the detector object because Charge attributes have changed
#     # detector.charge_list = collected_charge_list
#     # detector.charge = collected_charge_list      # TEMPORARY


# class Diffusion:
#     """TBW."""
#
#     def __init__(self, detector):
#         """TBW.
#
#         :param detector:
#         """
#         self.detector = detector
#
#         # Here is an image of all the last simulated CRs events on the CCD
#         self.pcmap_last = np.zeros((self.detector.row, self.detector.col))
#
#     # DIFFUSION
#     def janesick_diffusion_model(self, cluster):
#         """TBW."""
#         # Initial cloud diameter:
#         c_init = 0.0171 * (cluster.energy.value ** 1.75)
#
#         # 10 keV deposited by an X-ray photon resultsParticle a 1 um diameter charge (e-h) cloud
#         # CCD Advances For X - Ray Scientific Measurements In 1985,
#         # James Janesick et al.
#         # deltaE != cluster.number / u.electron * self.detector.ionization_energy / (1000 * u.eV)
#         # deltaE == kin. energy of an electron
#         # By analogy with high - energy electron beam interaction with silicon, one can approximate the
#         # energy / depth relationship as R = k * E**n , where k and n are numerical constants
#         # for the material and R is the penetration depth.
#
#         x_backside = self.detector.total_thickness      # um  # boundary of the detector backside
#         x_ff = self.detector.field_free_zone            # um # boundary of the field-free region near backside
#         x_p = self.detector.depletion_zone              # um # boundary of the depletion region near detector channel
#
#         # z position of charge generation event relative to the backside (>= 0)
#         x_a = x_backside + cluster.initial_position[2]
#
#         n_acceptor = 1e15 * u.cm ** (-3)
#
#         c_field_free = 0.0
#         # c_field = 0.0
#
#         if x_a == 0:
#             raise ValueError
#
#         # Cloud diameter after passing through a thin field-free region:
#         if 0 < x_a < x_ff:
#             c_field_free = 2 * x_ff * (1 - (x_a / x_ff) ** 2) ** 0.5
#
#         # Cloud diameter after passing through the field region:
#         # if x_ff <= x_a < x_ff + x_p:
#         c_field = (-2 * (5.1e-6 * (1e15 / n_acceptor.value) ** 0.5) ** 2 * log((x_a - x_ff) / x_p)) ** 0.5
#         c_field_max = 1.85e-5 * (1e15 / n_acceptor.value) ** 0.5
#         c_field = max(c_field, c_field_max)
#
#         # Charges already created inside CCD channel, not needed to diffuse them
#         if x_ff + x_p <= x_a:
#             c_field = 0.0
#
#         # Final cloud diameter: (um)
#         c_diameter = sqrt(c_init ** 2 + c_field_free ** 2 + c_field ** 2) ** 0.5 * u.um
#
#         return c_diameter
#
#     # DIFFUSION
#     def hiraga_diffusion_model(self, cluster):
#         """Spread the particle into the material.
#
#         Compute the density and size of the electronic cloud generated at each step.
#         """
#         pass
#         # eps_rel = 11.8                       # TODO: implement this in CCDDetector class
#         # eps_si = eps_rel * cds.eps0
#
#         # n_acceptor = 1e15 * u.cm**(-3)       # TODO: implement this in CCDDetector class
#
#         # voltage = cds.e * n_acceptor / (2 * eps_si) * l_dep**2    # V
#         # TODO: implement this in CCDDetector class
#         # assumptions made: V=0, dV/dz = 0 at z = ld
#         # voltage = 50 * u.V
#         # voltage = self.detector.bias_voltage
#
#         # depletion depth                      # TODO: implement this in CCDDetector class
#         # l_dep = sqrt(2 * eps_si * voltage / (cds.e * n_acceptor))   # cm
#         # l_dep = self.detector.depletion_zone
#
#         # critical field
#         # efield_crit = 1.01 * u.V/u.cm * self.detector.temperature ** 1.55        # V/cm
#         # For instance, typical value at T = 210 K are vs = 1.46e7 cm*sec−1, Ecrit = 4.4e3 V*cm−1 and α = 0.88
#         # Ecrit ~ 1e2 - 1e4 V*cm−1
#
#         # electron velocity saturation parameter
#         # sat = u.e * n_acceptor * l_dep / (eps_si * efield_crit)
#
#         # r_final =
#
#         # spreading across entire depletion region
#         # self.cfr = self.con * sqrt(self.sat + bound)
#
#         # return r_final
#
#     # ELECTRON COLLECTION
#     def gaussian_pixel_separation(self, cluster, sig_ac, sig_al):
#         """Compute the charge collection function.
#
#         Determine the number of electron collected by each pixel based on the
#         generated electronic cloud shape.
#
#         :param cluster:
#         :param float sig_ac: diameter of the resulting electronic cloud in the AC (across scan, vertical) dimension
#         :param float sig_al: diameter of the resulting electronic cloud in the AL (along scan, horizontal) dimension
#         """
#         self.pcmap_last[:, :] = 0
#         px = []
#         py = []
#
#         dx = (cluster.initial_position[0] - self.detector.pix_ver_size
#               * int(cluster.initial_position[0] / self.detector.pix_ver_size))
#         dy = (cluster.initial_position[1] - self.detector.pix_hor_size
#               * int(cluster.initial_position[1] / self.detector.pix_hor_size))
#
#         try:
#             int(4 * sig_ac / self.detector.pix_ver_size)  # WTF?
#         except ValueError:
#             print(sig_ac, cluster.number)
#
#         x_steps = int(4 * sig_ac / self.detector.pix_ver_size)
#
#         if x_steps > 49:  # WHY????
#             x_steps = 49
#         if x_steps < 1:
#             x_steps = 1
#
#         y_steps = int(4 * sig_al / self.detector.pix_hor_size)
#         if y_steps > 49:
#             y_steps = 49
#         if y_steps < 1:
#             y_steps = 1
#
#         for xi in np.arange(-(x_steps * self.detector.pix_ver_size + dx),
#                             ((x_steps + 1) * self.detector.pix_ver_size - dx),
#                             self.detector.pix_ver_size):
#
#             if sig_ac != 0:
#                 case1 = (xi + self.detector.pix_ver_size) / 1.41 / sig_ac
#                 case2 = xi / 1.41 / sig_ac
#             else:
#                 case1 = 0
#                 case2 = 0
#
#             px.append((erf(case1) - erf(case2)) / 2)
#
#         for yi in np.arange(-(y_steps * self.detector.pix_hor_size + dy),
#                             ((y_steps + 1) * self.detector.pix_hor_size - dy),
#                             self.detector.pix_hor_size):
#
#             if sig_al != 0:
#                 case1 = (yi + self.detector.pix_hor_size) / 1.41 / sig_al
#                 case2 = yi / 1.41 / sig_al
#             else:
#                 case1 = 0
#                 case2 = 0
#
#             py.append((erf(case1) - erf(case2)) / 2)
#
#         cx = 0
#
#         for ix in range(int(cluster.initial_position[0] / self.detector.pix_ver_size) - x_steps,
#                         int(cluster.initial_position[0] / self.detector.pix_ver_size) + x_steps + 1, 1):
#
#             cy = 0
#
#             for iy in range(int(cluster.initial_position[1] / self.detector.pix_hor_size) - y_steps,
#                             int(cluster.initial_position[1] / self.detector.pix_hor_size) + y_steps + 1, 1):
#
#                 if 0 <= ix < self.detector.row and 0 <= iy < self.detector.col:
#                     self.pcmap_last[ix, iy] += px[cx] * py[cy] * cluster.number
#
#                 cy += 1
#
#             cx += 1
#
#         # diff.pcmap_last = np.rint(diff.pcmap_last).astype(int)
#         # diff.total_charge_array += diff.pcmap_last
#
#         return self.pcmap_last



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Simple model to convert photon into photo-electrons inside detector."""
import typing as t
from pathlib import Path

import numpy as np
from typing_extensions import Literal

from pyxel.detectors import Detector
from pyxel.util import load_cropped_and_aligned_image


def apply_qe(array: np.ndarray, qe: t.Union[float, np.ndarray]) -> np.ndarray:
    """Apply quantum efficiency to an array.

    Parameters
    ----------
    array: np.ndarray
    qe: ndarray or float
        Quantum efficiency.

    Returns
    -------
    ndarray
    """
    return array * qe


def simple_conversion(
    detector: Detector, quantum_efficiency: t.Optional[float] = None
) -> None:
    """Generate charge from incident photon via photoelectric effect, simple model.

    Parameters
    ----------
    detector : Detector
        Pyxel Detector object.
    quantum_efficiency: float, optional
        Quantum efficiency.
    """
    if quantum_efficiency is None:
        final_qe = detector.characteristics.quantum_efficiency
    else:
        final_qe = quantum_efficiency

    if not 0 <= final_qe <= 1:
        raise ValueError("Quantum efficiency not between 0 and 1.")

    detector_charge = apply_qe(array=detector.photon.array, qe=final_qe)
    detector.charge.add_charge_array(detector_charge)


def conversion_with_qe_map(
    detector: Detector,
    filename: t.Union[str, Path],
    position: t.Tuple[int, int] = (0, 0),
    align: t.Optional[
        Literal["center", "top_left", "top_right", "bottom_left", "bottom_right"]
    ] = None,
) -> None:
    """Generate charge from incident photon via photoelectric effect, simple model with custom :term:`QE` map.

    Parameters
    ----------
    detector : Detector
        Pyxel Detector object.
    filename : str or Path
        File path.
    position: tuple
        Indices of starting row and column, used when fitting :term:`QE` map to detector.
    align: Literal
        Keyword to align the :term:`QE` map to detector. Can be any from:
        ("center", "top_left", "top_right", "bottom_left", "bottom_right")
    """
    geo = detector.geometry
    position_y, position_x = position

    # Load charge profile as numpy array.
    qe = load_cropped_and_aligned_image(
        shape=(geo.row, geo.col),
        filename=filename,
        position_x=position_x,
        position_y=position_y,
        align=align,
    )  # type: np.ndarray

    if not np.all((0 <= qe) & (qe <= 1)):
        raise ValueError("Quantum efficiency values not between 0 and 1.")

    detector_charge = apply_qe(array=detector.photon.array, qe=qe)
    detector.charge.add_charge_array(detector_charge)


# # TODO: Fix this
# # @validators.validate
# # @config.argument(name='', label='', units='', validate=)
# def monte_carlo_conversion(detector: Detector) -> None:
#     """Generate charge from incident photon via photoelectric effect, more exact, stochastic (Monte Carlo) model.
#
#     :param detector: Pyxel Detector object
#     """
#     logging.info("")
#
#     # detector.qe <= 1
#     # detector.eta <= 1
#     # if np.random.rand(size) <= detector.qe:
#     #     pass    # 1 e
#     # else:
#     #     pass
#     # if np.random.rand(size) <= detector.eta:
#     #     pass    # 1 e
#     # else:
#     #     pass
#     # TODO: random number for QE
#     # TODO: random number for eta
#     # TODO: energy threshold
#
#
# def random_pos(detector: Detector) -> None:
#     """Generate random position for photoelectric effect inside detector.
#
#     :param detector: Pyxel Detector object
#     """
#     # pos1 = detector.vert_dimension * np.random.random()
#     # pos2 = detector.horz_dimension * np.random.random()
#
#     # size = 0
#     # pos3 = -1 * detector.total_thickness * np.random.rand(size)
#     # return pos3
#     raise NotImplementedError



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Readout noise model."""
import typing as t

import numpy as np

from pyxel.detectors import CMOS, Detector
from pyxel.util import temporary_random_state


def create_noise(signal_2d: np.ndarray, std_deviation: float) -> np.ndarray:
    """Create noise to signal array using normal random distribution.

    Parameters
    ----------
    signal_2d : ndarray
    std_deviation : float

    Returns
    -------
    ndarray
    """
    sigma_2d = std_deviation * np.ones_like(signal_2d)
    noise_2d = np.random.normal(loc=signal_2d, scale=sigma_2d)

    return noise_2d


def create_noise_cmos(
    signal_2d: np.ndarray,
    readout_noise: float,
    readout_noise_std: float,
    charge_readout_sensitivity: float,
) -> np.ndarray:
    """Create noise to signal array for :term:`CMOS` detectors.

    Parameters
    ----------
    signal_2d : ndarray
    readout_noise : float
    readout_noise_std : float
    charge_readout_sensitivity : float

    Returns
    -------
    ndarray
    """
    sigma_2d = np.random.normal(
        loc=readout_noise * charge_readout_sensitivity,
        scale=readout_noise_std * charge_readout_sensitivity,
        size=signal_2d.shape,
    )

    noise_2d = np.random.normal(loc=signal_2d, scale=sigma_2d)

    return noise_2d


@temporary_random_state
def output_node_noise(
    detector: Detector,
    std_deviation: float,
    seed: t.Optional[int] = None,
) -> None:
    """Add noise to signal array of detector output node using normal random distribution.

    Parameters
    ----------
    detector: Detector
        Pyxel detector object.
    std_deviation: float
        Standard deviation. Unit: V
    seed: int, optional
        Random seed.

    Raises
    ------
    ValueError
        Raised if 'std_deviation' is negative
    """
    if std_deviation < 0.0:
        raise ValueError("'std_deviation' must be positive.")

    noise_2d = create_noise(
        signal_2d=np.asarray(detector.signal.array, dtype=np.float64),
        std_deviation=std_deviation,
    )  # type: np.ndarray

    detector.signal.array = noise_2d


@temporary_random_state
def output_node_noise_cmos(
    detector: CMOS,
    readout_noise: float,
    readout_noise_std: float,
    seed: t.Optional[int] = None,
) -> None:
    """Output node noise model for :term:`CMOS` detectors where readout is statistically independent for each pixel.

    Parameters
    ----------
    detector: CMOS
        Pyxel :term:`CMOS` object.
    readout_noise: float
        Mean readout noise for the array in units of electrons. Unit: electron
    readout_noise_std: float
        Readout noise standard deviation in units of electrons. Unit: electron
    seed: int, optional
        Random seed.

    Raises
    ------
    TypeError
        Raised if the 'detector' is not a :term:`CMOS` object.
    ValueError
        Raised if 'readout_noise_std' is negative.
    """
    if not isinstance(detector, CMOS):
        raise TypeError("Expecting a 'CMOS' detector object.")

    if readout_noise_std < 0.0:
        raise ValueError("'readout_noise_std' must be positive.")

    noise_2d = create_noise_cmos(
        signal_2d=np.asarray(detector.signal.array, dtype=np.float64),
        readout_noise=readout_noise,
        readout_noise_std=readout_noise_std,
        charge_readout_sensitivity=detector.characteristics.charge_to_volt_conversion,
    )  # type: np.ndarray

    detector.signal.array = noise_2d



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Pyxel photon generator models."""
import typing as t

import numpy as np
from typing_extensions import Literal

from pyxel.detectors import Detector


def rectangular(
    shape: t.Tuple[int, int],
    level: float,
    object_size: t.Optional[t.Sequence[int]] = None,
    object_center: t.Optional[t.Sequence[int]] = None,
) -> np.ndarray:
    """Calculate an image of a rectangular object.

    Parameters
    ----------
    shape: tuple
        Shape of the output array.
    level: float
        Flux of photon per pixel.
    object_size: list or tuple, optional
        List or tuple of length 2, integers defining the diameters of the rectangular object
        in vertical and horizontal directions.
    object_center: list or tuple, optional
        List or tuple of length 2, two integers (row and column number),
        defining the coordinates of the center of the rectangular object.

    Returns
    -------
    photon_array: ndarray
        Output numpy array.
    """
    if not object_size:
        raise ValueError(
            "object_size argument should be defined for illumination model"
        )
    if object_size and not len(object_size) == 2:
        raise ValueError("Object size should be a sequence of length 2!")
    if object_center and not len(object_center) == 2:
        raise ValueError("Object size should be a sequence of length 2!")

    photon_array = np.zeros(shape, dtype=float)
    if object_center is not None:
        if not (
            (0 <= object_center[0] <= shape[0]) and (0 <= object_center[1] <= shape[1])
        ):
            raise ValueError('Argument "object_center" should be inside Photon array.')
    else:
        object_center = [int(shape[0] / 2), int(shape[1] / 2)]
    p = object_center[0] - int(object_size[0] / 2)
    q = object_center[1] - int(object_size[1] / 2)
    p0 = int(np.clip(p, a_min=0, a_max=shape[0]))
    q0 = int(np.clip(q, a_min=0, a_max=shape[1]))
    photon_array[slice(p0, p + object_size[0]), slice(q0, q + object_size[1])] = level

    return photon_array


def elliptic(
    shape: t.Tuple[int, int],
    level: float,
    object_size: t.Optional[t.Sequence[int]] = None,
    object_center: t.Optional[t.Sequence[int]] = None,
) -> np.ndarray:
    """Calculate an image of an elliptic object.

    Parameters
    ----------
    shape: tuple
        Shape of the output array.
    level: float
        Flux of photon per pixel.
    object_size: list or tuple, optional
        List or tuple of length 2, integers defining the diameters of the elliptic object
        in vertical and horizontal directions.
    object_center: list or tuple, optional
        List or tuple of length 2, two integers (row and column number),
        defining the coordinates of the center of the elliptic object.

    Returns
    -------
    photon_array: ndarray
        Output numpy array.
    """
    if not object_size:
        raise ValueError(
            "object_size argument should be defined for illumination model"
        )
    if object_size and not len(object_size) == 2:
        raise ValueError("Object size should be a sequence of length 2!")
    if object_center and not len(object_center) == 2:
        raise ValueError("Object size should be a sequence of length 2!")

    photon_array = np.zeros(shape, dtype=float)
    if object_center is not None:
        if not (
            (0 <= object_center[0] <= shape[0]) and (0 <= object_center[1] <= shape[1])
        ):
            raise ValueError('Argument "object_center" should be inside Photon array.')
    else:
        object_center = [int(shape[0] / 2), int(shape[1] / 2)]
    y, x = np.ogrid[: shape[0], : shape[1]]
    dist_from_center = np.sqrt(
        ((x - object_center[1]) / float(object_size[1] / 2)) ** 2
        + ((y - object_center[0]) / float(object_size[0] / 2)) ** 2
    )
    photon_array[dist_from_center < 1] = level
    return photon_array


def calculate_illumination(
    shape: t.Tuple[int, int],
    level: float,
    option: str = "uniform",
    object_size: t.Optional[t.Sequence[int]] = None,
    object_center: t.Optional[t.Sequence[int]] = None,
) -> np.ndarray:
    """Calculate the array of photons uniformly over the entire array or over a object.

    Parameters
    ----------
    shape: tuple
        Shape of the output array.
    level: float
        Flux of photon per pixel.
    option: str{'uniform', 'elliptic_hole', 'rectangular_hole'}
        A string indicating the type of illumination:
        - ``uniform``
           Uniformly fill the entire array with photon. (Default)
        - ``elliptic_hole``
           Mask with elliptic object.
        - ``rectangular_hole``
           Mask with rectangular object.
    object_size: list or tuple, optional
        List or tuple of length 2, integers defining the diameters of the elliptic or rectangular object
        in vertical and horizontal directions.
    object_center: list or tuple, optional
        List or tuple of length 2, two integers (row and column number),
        defining the coordinates of the center of the elliptic or rectangular object.

    Returns
    -------
    photon_array: ndarray
        Output numpy array.
    """
    if option == "uniform":
        photon_array = np.ones(shape, dtype=float) * level
    elif option == "rectangular":
        photon_array = rectangular(
            shape=shape,
            object_size=object_size,
            object_center=object_center,
            level=level,
        )
    elif option == "elliptic":
        photon_array = elliptic(
            shape=shape,
            object_size=object_size,
            object_center=object_center,
            level=level,
        )
    else:
        raise NotImplementedError

    return photon_array


def illumination(
    detector: Detector,
    level: float,
    option: Literal["uniform", "rectangular_hole", "elliptic_hole"] = "uniform",
    object_size: t.Optional[t.Sequence[int]] = None,
    object_center: t.Optional[t.Sequence[int]] = None,
    time_scale: float = 1.0,
) -> None:
    """Generate photon uniformly over the entire array or over an elliptic or rectangular object.

    detector: Detector
        Pyxel Detector object.
    level: float
        Flux of photon per pixel.
    option: str
        A string indicating the type of illumination:
        - ``uniform``
        Uniformly fill the entire array with photon. (Default)
        - ``elliptic``
        Mask with elliptic object.
        - ``rectangular``
        Mask with rectangular object.
    object_size: list or tuple, optional
        List or tuple of length 2, integers defining the diameters of the elliptic or rectangular object
        in vertical and horizontal directions.
    object_center: list or tuple, optional
        List or tuple of length 2, two integers (row and column number),
        defining the coordinates of the center of the elliptic or rectangular object.
    time_scale: float
        Time scale of the photon flux, default is 1 second. 0.001 would be ms.
    """
    shape = (detector.geometry.row, detector.geometry.col)

    photon_array = calculate_illumination(
        shape=shape,
        level=level,
        option=option,
        object_size=object_size,
        object_center=object_center,
    )

    photon_array = photon_array * (detector.time_step / time_scale)

    try:
        detector.photon.array += photon_array
    except ValueError as ex:
        raise ValueError("Shapes of arrays do not match") from ex



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.
#
#
"""Utility functions for images."""
import typing as t
from enum import Enum
from functools import lru_cache
from pathlib import Path

import numpy as np
from typing_extensions import Literal

from pyxel.inputs import load_image


class Alignment(Enum):
    """Alignment class."""

    center = "center"
    top_left = "top_left"
    top_right = "top_right"
    bottom_left = "bottom_left"
    bottom_right = "bottom_right"


def _set_relative_position(
    array_x: int,
    array_y: int,
    output_x: int,
    output_y: int,
    alignment: Alignment,
) -> t.Tuple[int, int]:
    """Calculate relative position of (0, 0) pixels for two different array shapes based on desired alignment.

    Parameters
    ----------
    output_y: int
    output_x: int
    array_y: int
    array_x: int
    alignment: Alignment

    Returns
    -------
    tuple
    """
    if alignment == Alignment.center:
        return int((output_y - array_y) / 2), int((output_x - array_x) / 2)
    elif alignment == Alignment.top_left:
        return 0, 0
    elif alignment == Alignment.top_right:
        return 0, output_x - array_x
    elif alignment == Alignment.bottom_left:
        return output_y - array_y, 0
    elif alignment == Alignment.bottom_right:
        return output_y - array_y, output_x - array_x
    else:
        raise NotImplementedError


def fit_into_array(
    array: np.ndarray,
    output_shape: t.Tuple[int, ...],
    relative_position: t.Tuple[int, int] = (0, 0),
    align: t.Optional[
        Literal["center", "top_left", "top_right", "bottom_left", "bottom_right"]
    ] = None,
    allow_smaller_array: bool = True,
) -> np.ndarray:
    """Fit input array into an output array of specified output shape.

    Input array can be either larger or smaller than output array. In the first case the input array will be cropped.
    The relative position between the arrays in the coordinate system is specified with argument relative_position.
    It is a tuple with coordinates (Y,X).
    User can use this argument to specify the position of input array values in the output array.
    If arrays are to be aligned in the center or one of the corners,
    it is also possible with argument align and passing a location keyword.
    User can turn on that smaller arrays than output shape are not allowed by setting allow_smaller_array to False.

    Parameters
    ----------
    array: ndarray
    output_shape: tuple
    relative_position: tuple
    align: {'center', 'top_left', 'top_right', 'bottom_left', 'bottom_right'}, default: None
    allow_smaller_array: bool

    Returns
    -------
    output: ndarray
    """

    array_y, array_x = array.shape
    output = np.zeros(output_shape)
    output_y, output_x = output_shape

    if not allow_smaller_array:
        if array_y < output_y or array_x < output_y:
            raise ValueError("Input array too small to fit into the desired shape!.")

    if align:
        relative_position = _set_relative_position(
            array_x=array_x,
            array_y=array_y,
            output_x=output_x,
            output_y=output_y,
            alignment=Alignment(align),
        )

    array_y_coordinates = np.array(
        range(relative_position[0], relative_position[0] + array_y)
    )
    array_x_coordinates = np.array(
        range(relative_position[1], relative_position[1] + array_x)
    )
    output_y_coordinates = np.array(range(output_y))
    output_x_coordinates = np.array(range(output_x))

    overlap_y = np.intersect1d(array_y_coordinates, output_y_coordinates)
    overlap_x = np.intersect1d(array_x_coordinates, output_x_coordinates)

    if overlap_y.size == 0 and overlap_x.size == 0:
        raise ValueError("No overlap of array and target in Y and X dimension.")
    elif overlap_y.size == 0:
        raise ValueError("No overlap of array and target in Y dimension.")
    elif overlap_x.size == 0:
        raise ValueError("No overlap of array and target in X dimension.")

    cropped_array = array[
        slice(
            overlap_y[0] - relative_position[0],
            overlap_y[-1] + 1 - relative_position[0],
        ),
        slice(
            overlap_x[0] - relative_position[1],
            overlap_x[-1] + 1 - relative_position[1],
        ),
    ]

    output[
        slice(overlap_y[0], overlap_y[-1] + 1), slice(overlap_x[0], overlap_x[-1] + 1)
    ] = cropped_array

    return output


@lru_cache(maxsize=128)  # One must add parameter 'maxsize' for Python 3.7
def load_cropped_and_aligned_image(
    shape: t.Tuple[int, ...],
    filename: t.Union[str, Path],
    position_x: int = 0,
    position_y: int = 0,
    align: t.Optional[
        Literal["center", "top_left", "top_right", "bottom_left", "bottom_right"]
    ] = None,
    allow_smaller_array: bool = True,
) -> np.ndarray:
    """Load image from file and fit to detector shape.

    Parameters
    ----------
    shape: tuple
        Detector shape.
    filename: str
        Path to image file.
    position_x: tuple
        Index of starting column, used when fitting image to detector.
    position_y: tuple
        Index of starting row, used when fitting image to detector.
    align: Literal
        Keyword to align the image to detector. Can be any from:
        ("center", "top_left", "top_right", "bottom_left", "bottom_right")

    Returns
    -------
    cropped_and_aligned_image: ndarray
    """
    # Load 2d image (which can be smaller or
    #                         larger in dimensions than detector imaging area)
    image = load_image(filename)

    cropped_and_aligned_image = fit_into_array(
        array=image,
        output_shape=shape,
        relative_position=(position_y, position_x),
        align=align,
        allow_smaller_array=allow_smaller_array,
    )  # type: np.ndarray

    return cropped_and_aligned_image



"""Unittests for the 'ModelFitting' class."""
#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

from pathlib import Path

import numpy as np
import pytest

import pyxel
from pyxel.calibration import Calibration
from pyxel.calibration.fitting import ModelFitting
from pyxel.calibration.util import CalibrationMode
from pyxel.detectors import CCD
from pyxel.pipelines import DetectionPipeline, Processor
from pyxel.pipelines.processor import ResultType

try:
    import pygmo as pg

    WITH_PYGMO = True
except ImportError:
    WITH_PYGMO = False


def configure(mf: ModelFitting, sim: pyxel.Configuration) -> None:
    """TBW."""
    assert sim.calibration is not None

    pg.set_global_rng_seed(sim.calibration.pygmo_seed)
    np.random.seed(sim.calibration.pygmo_seed)

    mf.configure(
        calibration_mode=sim.calibration.calibration_mode,
        generations=sim.calibration.algorithm.generations,
        population_size=sim.calibration.algorithm.population_size,
        simulation_output=sim.calibration.result_type,
        fitness_func=sim.calibration.fitness_function,
        target_output=sim.calibration.target_data_path,
        target_fit_range=sim.calibration.target_fit_range,
        out_fit_range=sim.calibration.result_fit_range,
        weights_from_file=sim.calibration.weights_from_file,
        weights=sim.calibration.weights,
        file_path=None,
    )


@pytest.mark.skipif(not WITH_PYGMO, reason="Package 'pygmo' is not installed.")
@pytest.mark.parametrize("yaml_file", ["tests/data/calibrate.yaml"])
def test_configure_params(yaml_file):
    """Test"""
    cfg = pyxel.load(yaml_file)
    detector = cfg.ccd_detector
    pipeline = cfg.pipeline
    processor = Processor(detector, pipeline)
    calibration = cfg.calibration
    mf = ModelFitting(
        processor=processor,
        variables=calibration.parameters,
        readout=calibration.readout,
    )

    if not isinstance(mf.processor, Processor):
        raise ValueError

    configure(mf, cfg)

    assert mf.calibration_mode == CalibrationMode.Pipeline
    assert mf.sim_fit_range == (
        slice(None, None, None),
        slice(0, 4, None),
        slice(2, 4, None),
    )
    assert mf.targ_fit_range == (slice(1, 5, None), slice(0, 2, None))
    assert mf.sim_output == ResultType.Image


@pytest.mark.skipif(not WITH_PYGMO, reason="Package 'pygmo' is not installed.")
@pytest.mark.parametrize("yaml", ["tests/data/calibrate_fits.yaml"])
def test_configure_fits_target(yaml):
    """Test"""
    cfg = pyxel.load(yaml)
    detector = cfg.ccd_detector
    pipeline = cfg.pipeline
    processor = Processor(detector, pipeline)
    calibration = cfg.calibration
    mf = ModelFitting(
        processor=processor,
        variables=calibration.parameters,
        readout=calibration.readout,
    )
    configure(mf, cfg)
    assert mf.sim_fit_range == (
        slice(None, None, None),
        slice(0, 4, None),
        slice(2, 4, None),
    )
    assert mf.targ_fit_range == (slice(1, 5, None), slice(0, 2, None))
    assert mf.sim_output == ResultType.Image
    expected = np.array(
        [
            [4173.6434, 4203.6883],
            [4468.6537, 4517.2588],
            [4683.7958, 4594.2287],
            [4520.4915, 4315.9494],
        ]
    )
    np.testing.assert_array_equal(
        np.around(mf.all_target_data[0], decimals=4), np.around(expected, decimals=4)
    )


@pytest.mark.skipif(not WITH_PYGMO, reason="Package 'pygmo' is not installed.")
@pytest.mark.parametrize(
    "yaml", ["tests/data/calibrate.yaml", "tests/data/calibrate_fits.yaml"]
)
def test_boundaries(yaml):
    """Test"""
    cfg = pyxel.load(yaml)
    detector = cfg.ccd_detector
    pipeline = cfg.pipeline
    processor = Processor(detector, pipeline)
    calibration = cfg.calibration
    mf = ModelFitting(
        processor=processor,
        variables=calibration.parameters,
        readout=calibration.readout,
    )

    configure(mf, cfg)

    lbd_expected = [1.0, -3.0, -3.0, -2.0, -2.0, 0.0, 10.0]
    ubd_expected = [10.0, 0.3010299956639812, 0.3010299956639812, 1.0, 1.0, 1.0, 200.0]
    assert mf.lbd == lbd_expected
    assert mf.ubd == ubd_expected
    ll, uu = mf.get_bounds()
    assert ll == lbd_expected
    assert uu == ubd_expected


@pytest.mark.skipif(not WITH_PYGMO, reason="Package 'pygmo' is not installed.")
@pytest.mark.parametrize(
    "simulated_data, target_data, expected_fitness",
    [
        # (231, 231, 0.0),
        # (231, 145, 86.0),
        # (2.31, 1.45, 0.8600000000000001),
        # (2.0, 1, 1.0),
        (np.array([1, 9, 45, 548, 2, 2]), np.array([1, 9, 45, 548, 2, 2]), 0.0),
        (np.array([1, 9, 45, 548, 2, 2]), np.array([1, 3, 56, 21, 235, 11]), 786.0),
        (
            np.array([[1362.0, 1378.0], [1308.0, 1309.0]]),
            np.array([[1362.0, 1378.0], [1308.0, 1309.0]]),
            0.0,
        ),
        (
            np.array([[1362.0, 1378.0], [1308.0, 1309.0]]),
            np.array([[1462.0, 1368.0], [1508.0, 1399.0]]),
            400.0,
        ),
    ],
)
def test_calculate_fitness(simulated_data, target_data, expected_fitness):
    """Test"""
    cfg = pyxel.load("tests/data/calibrate.yaml")
    detector = cfg.ccd_detector
    pipeline = cfg.pipeline
    processor = Processor(detector, pipeline)
    calibration = cfg.calibration
    mf = ModelFitting(
        processor=processor,
        variables=calibration.parameters,
        readout=calibration.readout,
    )
    configure(mf, cfg)
    fitness = mf.calculate_fitness(simulated_data, target_data)
    assert fitness == expected_fitness
    print("fitness: ", fitness)


@pytest.mark.skipif(not WITH_PYGMO, reason="Package 'pygmo' is not installed.")
@pytest.mark.parametrize(
    "yaml, factor, expected_fitness",
    [
        (Path("tests/data/calibrate_weighting.yaml"), 1, 0.0),
        (Path("tests/data/calibrate_weighting.yaml"), 2, 1287593.3479011902),
        (Path("tests/data/calibrate_weighting.yaml"), 3, 2575186.695802381),
    ],
)
def test_weighting(yaml, factor, expected_fitness):
    """Test"""
    cfg = pyxel.load(yaml)
    detector = cfg.ccd_detector
    pipeline = cfg.pipeline
    processor = Processor(detector, pipeline)
    calibration = cfg.calibration
    mf = ModelFitting(
        processor=processor,
        variables=calibration.parameters,
        readout=calibration.readout,
    )
    configure(mf, cfg)
    fitness = mf.calculate_fitness(
        mf.all_target_data[0] * factor, mf.all_target_data[0]
    )
    assert fitness == expected_fitness
    print("fitness: ", fitness)


def custom_fitness_func(simulated, target, weighting=None):
    """Custom fitness func for testing"""
    return np.sum(target * 2 - simulated / 2 + 1.0)


@pytest.mark.skipif(not WITH_PYGMO, reason="Package 'pygmo' is not installed.")
@pytest.mark.parametrize(
    "yaml, simulated, target, weighting",
    [
        (
            "tests/data/calibrate_custom_fitness.yaml",
            np.array([1.0]),
            np.array([2.0]),
            np.array([4.5]),
        ),
        (
            "tests/data/calibrate_custom_fitness.yaml",
            np.array([1.0, 2.0, 3.0]),
            np.array([2.0, 5.0, 6.0]),
            np.array([26.0]),
        ),
        (
            "tests/data/calibrate_least_squares.yaml",
            np.array([2.0]),
            np.array([4.0]),
            np.array([4.0]),
        ),
    ],
)
def test_custom_fitness(yaml, simulated, target, weighting):
    """Test"""
    cfg = pyxel.load(yaml)
    assert isinstance(cfg, pyxel.Configuration)

    detector = cfg.ccd_detector
    assert isinstance(detector, CCD)

    pipeline = cfg.pipeline
    assert isinstance(pipeline, DetectionPipeline)

    processor = Processor(detector, pipeline)
    assert isinstance(processor, Processor)

    calibration = cfg.calibration
    assert isinstance(calibration, Calibration)

    assert cfg.calibration is not None

    mf = ModelFitting(
        processor=processor,
        variables=calibration.parameters,
        readout=calibration.readout,
    )
    configure(mf=mf, sim=cfg)

    fitness = mf.calculate_fitness(simulated, target)
    assert fitness == weighting
    print("fitness: ", fitness)


@pytest.mark.skipif(not WITH_PYGMO, reason="Package 'pygmo' is not installed.")
@pytest.mark.parametrize(
    "yaml, parameter, expected_fitness",
    [
        (
            "tests/data/calibrate_models.yaml",
            np.array(
                [
                    1.0,
                    0.5,
                    1.5,
                    -2.0,
                    -3.0,
                    4.5,
                    -4.0,
                    1.0,
                    0.5,
                    -3.5,
                    2.0,
                    -3.0,
                    -4.0,
                    0.5,
                    1.0,
                    100.0,
                ]
            ),
            88455.49014776475,
        )
    ],
)
def test_fitness(yaml, parameter, expected_fitness):
    """Test"""
    cfg = pyxel.load(yaml)

    detector = cfg.ccd_detector
    pipeline = cfg.pipeline

    processor = Processor(detector, pipeline)
    calibration = cfg.calibration

    mf = ModelFitting(
        processor=processor,
        variables=calibration.parameters,
        readout=calibration.readout,
    )

    configure(mf, cfg)

    overall_fitness = mf.fitness(parameter)
    assert overall_fitness[0] == expected_fitness

    print("fitness: ", overall_fitness[0])


@pytest.mark.skipif(not WITH_PYGMO, reason="Package 'pygmo' is not installed.")
@pytest.mark.parametrize(
    "yaml, parameter, expected_array",
    [
        (
            "tests/data/calibrate_models.yaml",
            np.arange(16.0),
            np.array(
                [
                    0.0,
                    10.0,
                    100.0,
                    1000.0,
                    10000.0,
                    5.0,
                    6.0,
                    7.0,
                    8.0,
                    1.0e09,
                    1.0e10,
                    1.0e11,
                    1.0e12,
                    13.0,
                    1.0e14,
                    15.0,
                ]
            ),
        )
    ],
)
def test_split_and_update(yaml, parameter, expected_array):
    """Test"""
    cfg = pyxel.load(yaml)
    detector = cfg.ccd_detector
    pipeline = cfg.pipeline
    processor = Processor(detector, pipeline)
    calibration = cfg.calibration
    mf = ModelFitting(
        processor=processor,
        variables=calibration.parameters,
        readout=calibration.readout,
    )
    configure(mf, cfg)
    array = mf.convert_to_parameters(parameter)
    np.testing.assert_array_equal(array, expected_array)


@pytest.mark.skipif(not WITH_PYGMO, reason="Package 'pygmo' is not installed.")
@pytest.mark.parametrize(
    "yaml, param_array",
    [
        (
            "tests/data/calibrate_models.yaml",
            np.array(
                [
                    1.0,
                    10.0,
                    100.0,
                    1000.0,
                    10000.0,
                    5.0,
                    6.0,
                    7.0,
                    8.0,
                    1.0e09,
                    1.0e10,
                    1.0e11,
                    1.0e12,
                    13.0,
                    1.0e14,
                    150,
                ]
            ),
        )
    ],
)
def test_detector_and_model_update(yaml: str, param_array: np.ndarray):
    """Test"""
    cfg = pyxel.load(yaml)

    detector = cfg.ccd_detector
    pipeline = cfg.pipeline

    processor = Processor(detector, pipeline)
    calibration = cfg.calibration
    mf = ModelFitting(
        processor=processor,
        variables=calibration.parameters,
        readout=calibration.readout,
    )
    configure(mf=mf, sim=cfg)
    mf.processor = mf.update_processor(param_array, processor)

    assert mf.processor.pipeline.charge_transfer is not None
    assert mf.processor.pipeline.charge_measurement is not None

    attributes = [
        mf.processor.detector.characteristics.pre_amplification,
        mf.processor.pipeline.charge_transfer.models[0].arguments["trap_release_times"],
        mf.processor.pipeline.charge_transfer.models[0].arguments["trap_densities"],
        mf.processor.pipeline.charge_transfer.models[0].arguments["sigma"],
        mf.processor.pipeline.charge_transfer.models[0].arguments["beta"],
        mf.processor.pipeline.charge_measurement.models[1].arguments["std_deviation"],
        mf.processor.detector.environment.temperature,
    ]
    a = 0
    for idx, attr in enumerate(attributes):
        if isinstance(attr, np.ndarray):
            b = len(attr)
            np.testing.assert_array_equal(attr, param_array[a : a + b])
        else:
            b = 1
            assert attr == param_array[a]
        a += b



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""TBW."""

import astropy.constants as const
import astropy.units as u
import numpy as np

from pyxel.detectors import MKID


def apply_dead_time_filter(phase_2d: np.ndarray, maximum_count: float) -> np.ndarray:
    """Apply dead time filter.

    Parameters
    ----------
    phase_2d : ndarray
    maximum_count : float

    Returns
    -------
    ndarray
    """
    phase_clipped_2d = np.clip(phase_2d, a_min=None, a_max=maximum_count)

    return phase_clipped_2d


# TODO: more documentation (Enrico). See #324.
def dead_time_filter(
    detector: MKID,
    tau_0: float = 4.4 * 1.0e-7,
    n_0: float = 1.72 * 1.0e10,
    t_c: float = 1.26,
    v: float = 30.0,
    t_op: float = 0.3,
    tau_pb: float = 2.8 * 1.0e-10,
    tau_esc: float = 1.4 * 1.0e-10,
    tau_sat: float = 1.0e-3,
) -> None:
    """Dead time filter.

    The underlying physics of this model is described in :cite:p:`PhysRevB.104.L180506`;
    more information can be found on the website :cite:p:`Mazin`.

    Parameters
    ----------
    detector: MKID
        Pyxel Detector :term:`MKID` object.
    tau_0 : float
        Material dependent characteristic time for the electron-phonon coupling. Unit: s
    n_0 : float
        Material dependent single spin density of states at
        the Fermi-level. Unit: um^-3 eV^-1
    t_c : float
        Material dependent critical temperature. Unit: K
    v : float
        Superconducting volume. Unit: um^3
    t_op : float
        Temperature. Unit: K
    tau_pb : float
        Phonon pair-breaking time. Unit: s
    tau_esc : float
        Phonon escape time. Unit: s
    tau_sat : float
        Saturation time. Unit: s
    """
    # Validation phase
    if not isinstance(detector, MKID):
        # Later, this will be checked in when YAML configuration file is parsed
        raise TypeError("Expecting an `MKID` object for 'detector'.")

    # Boltzmann's constant [eV K^-1]
    boltzmann_cst: float = const.k_B.to(u.eV / u.K).value

    # Compute superconducting gap energy
    delta = 1.76 * boltzmann_cst * t_c  # type: float

    # Compute number of quasiparticles in a superconducting volume V
    # TODO: check that T << (delta / boltzmann_cst)
    n_qp = (
        2.0
        * v
        * n_0
        * np.sqrt(2.0 * np.pi * boltzmann_cst * t_op * delta)
        * np.exp(-delta / (boltzmann_cst * t_op))
    )  # type: float

    # Compute recombination constant
    recombination_cst = (
        tau_0 * n_0 * (boltzmann_cst * t_c) ** 3 / (2.0 * delta ** 2)
    )  # type: float

    # Compute intrinsic quasiparticle lifetime with respect to recombination
    tau_qp = v / (recombination_cst * n_qp)  # type: float

    # Compute apparent quasi-particle lifetime, without saturation lifetime
    # tau_apparent = 1. / (2. / (tau_qp * (1. + (char.tau_esc / char.tau_pb))))

    # Compute apparent quasi-particle lifetime, including saturation lifetime
    tau_apparent_sat = 1.0 / (
        2.0 / (tau_qp * (1.0 + (tau_esc / tau_pb))) + (1.0 / tau_sat)
    )  # type: float

    dead_time = tau_apparent_sat

    phase_2d = apply_dead_time_filter(
        phase_2d=detector.phase.array,
        maximum_count=1.0 / dead_time,
    )

    detector.phase.array = phase_2d



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""
Persistence module for the PyXel simulation.

This module is used in charge_collection

-------------------------------------------------------------------------------

+--------------+----------------------------------+---------------------------+
| Author       | Name                             | Creation                  |
+--------------+----------------------------------+---------------------------+
| You          | charge_collection                | 05/20/2021                |
+--------------+----------------------------------+---------------------------+

+-----------------+-------------------------------------+---------------------+
| Contributor     | Name                                | Creation            |
+-----------------+-------------------------------------+---------------------+
| Name            | filename                            | -                   |
+-----------------+-------------------------------------+---------------------+

This module can be found in pyxel/models/charge_collection/persistence.py

Algorithm
=========

The persistence model take as input the total number of detrapped charges
(persistence map with a 10000 seconds soak at more than the FWC) and a map
which is the previous one divided by full well capacity. This gives the total
amount of trap per charge.

At each iteration of the pipeline, the model  will compute the amount of trapped charges in this iteration, add it
to the memory of the detector and then will remove this amount from the pixel array.


Code example
============

.. code-block:: python

    # To access the memory of the detector with the persistence map

    detector.persistence

Model reference in the YAML config file
=======================================

.. code-block:: yaml

pipeline:

  # Persistence model based on MOONS detector (H4RG) measurements
  - name: persistence
    func: pyxel.models.charge_collection.persistence
    enabled: true
    arguments:
      trap_time_constants: [1, 10, 100, 1000, 10000]
      trap_densities_filename: data/fits/20210408121614_20210128_ENG20370_AUTOCHAR-Persistence_FitTrapDensityMap.fits
      trap_capacities_filename: data/fits/20210408093114_20210128_ENG20370_AUTOCHAR-Persistence_FitMaximumTrapMap.fits
      trap_proportions: [0.307, 0.175, 0.188, 0.136, 0.194]

Useful links
============

Persistence paper from S. Tulloch
https://arxiv.org/abs/1908.06469

TO DO:
   - Add temperature dependency
   - Add default -flat?- trap maps

"""

import typing as t
from pathlib import Path

import numba
import numpy as np
from typing_extensions import Literal

from pyxel.data_structure import Persistence, SimplePersistence
from pyxel.detectors import CMOS
from pyxel.util import load_cropped_and_aligned_image


def simple_persistence(
    detector: CMOS,
    trap_time_constants: t.Sequence[float],
    trap_densities: t.Sequence[float],
    trap_capacities: t.Optional[t.Sequence[float]] = None,
) -> None:
    """Apply simple persistence model.

    Parameters
    ----------
    detector: CMOS
        Pyxel CMOS object.
    trap_time_constants: sequence of floats
        List of trap time constants.
    trap_densities: sequence of floats
        List of trap densities.
    trap_capacities: sequence of floats, optional
        List of trap capacities.
    """

    if not isinstance(detector, CMOS):
        raise TypeError("Expecting a CMOS object for detector.")

    if not len(trap_time_constants) == len(trap_densities):
        raise ValueError(
            "Expecting same number of elements for parameters 'trap_time_constants'"
            "and 'trap_densities'"
        )

    if len(trap_time_constants) == 0:
        raise ValueError(
            "Expecting at least one 'trap_time_constants' and 'trap_densities'"
        )

    if trap_capacities is not None and not len(trap_capacities) == len(
        trap_time_constants
    ):
        raise ValueError(
            "Expecting same number of elements for parameters 'trap_capacities'"
            "and 'trap_time_constants'"
        )

    if not detector.has_persistence():
        detector.persistence = SimplePersistence(
            trap_time_constants=trap_time_constants,
            trap_densities=trap_densities,
            geometry=detector.pixel.shape,
        )

    if trap_capacities is None:
        fwc = None
    else:
        fwc = np.array(trap_capacities)

    new_pixel_array, new_all_trapped_charge = compute_simple_persistence(
        pixel_array=detector.pixel.array,
        all_trapped_charge=detector.persistence.trapped_charge_array,
        trap_densities=np.array(trap_densities),
        trap_time_constants=np.array(trap_time_constants),
        trap_capacities=fwc,
        delta_t=detector.time_step,
    )

    detector.pixel.array = new_pixel_array
    detector.persistence.trapped_charge_array = new_all_trapped_charge


@numba.njit(fastmath=True)
def compute_simple_persistence(
    pixel_array: np.ndarray,
    all_trapped_charge: np.ndarray,
    trap_densities: np.ndarray,
    trap_time_constants: np.ndarray,
    delta_t: float,
    trap_capacities: t.Optional[np.ndarray] = None,
) -> t.Tuple[np.ndarray, np.ndarray]:
    """Compute simple persistence.

    Parameters
    ----------
    pixel_array: ndarray
    all_trapped_charge: ndarray
    trap_densities: ndarray
    trap_time_constants: ndarray
    trap_capacities: ndarray, optional
    delta_t: float

    Returns
    -------
    tuple of ndarray
    """
    pixel_start = pixel_array.copy()

    for i, trapped_charge in enumerate(all_trapped_charge):
        # Computer trapped charge for this increment of time
        # Time factor is the integration time divided by the time constant (1, 10, 100, 1000, 10000)
        time_factor = delta_t / trap_time_constants[i]

        # Compute trapped charges
        available_traps = trap_densities[i] * pixel_array
        empty_traps = available_traps - trapped_charge
        diff = time_factor * empty_traps

        # maximum you can release is trapped_charge, max you can add is empty traps
        diff = clip_diff(
            diff=diff, trapped_charge=trapped_charge, empty_traps=empty_traps
        )

        trapped_charge += diff

        pixel_array -= diff

        all_trapped_charge[i] = trapped_charge

    pixel_diff = pixel_array - pixel_start

    for i, trapped_charge in enumerate(all_trapped_charge):

        if trap_capacities is None:
            fwc = None
        else:
            fwc = trap_capacities[i] * np.ones(trapped_charge.shape)

        densities = trap_densities[i] * np.ones(trapped_charge.shape)
        available_traps = pixel_array * densities

        trapped_charge_clipped, output_pixel = clip_trapped_charge(
            trapped_charge=trapped_charge,
            pixel=pixel_array,
            available_traps=available_traps,
            pixel_diff=pixel_diff,
            trap_capacities=fwc,
        )
        all_trapped_charge[i] = trapped_charge_clipped

    return output_pixel, all_trapped_charge


def persistence(
    detector: CMOS,
    trap_time_constants: t.Sequence[float],
    trap_proportions: t.Sequence[float],
    trap_densities_filename: t.Union[Path, str],
    trap_capacities_filename: t.Optional[t.Union[Path, str]] = None,
    trap_densities_position: t.Tuple[int, int] = (0, 0),
    trap_densities_align: t.Optional[
        Literal["center", "top_left", "top_right", "bottom_left", "bottom_right"]
    ] = None,
    trap_capacities_position: t.Tuple[int, int] = (0, 0),
    trap_capacities_align: t.Optional[
        Literal["center", "top_left", "top_right", "bottom_left", "bottom_right"]
    ] = None,
) -> None:
    """Apply persistence model.

    Parameters
    ----------
    detector: CMOS
        Pyxel CMOS object.
    trap_time_constants: sequence of floats
        A list of trap time constants.
    trap_proportions: sequence of floats
        A list of trap proportions.
    trap_densities_filename: path or str
        Path to densities file.
    trap_capacities_filename: path or str
        Path to capacities file.
    trap_densities_position: tuple of int
        Indices of starting row and column, used when fitting densities to detector.
    trap_densities_align: literal
        Keyword to align the densities to detector. Can be any from:
        ("center", "top_left", "top_right", "bottom_left", "bottom_right")
    trap_capacities_position: tuple of int
        Indices of starting row and column, used when fitting capacities to detector.
    trap_capacities_align: literal
        Keyword to align the capacities to detector. Can be any from:
        ("center", "top_left", "top_right", "bottom_left", "bottom_right")
    """

    if not isinstance(detector, CMOS):
        raise TypeError("Expecting a CMOS object for detector.")

    if not len(trap_time_constants) == len(trap_proportions):
        raise ValueError(
            "Expecting same number of elements for parameters 'trap_time_constants'"
            "and 'trap_proportions'"
        )

    if len(trap_time_constants) == 0:
        raise ValueError(
            "Expecting at least one 'trap_time_constants' and 'trap_proportions'"
        )

    # If the file for trap density is correct open it and use it
    # otherwise I need to define a default trap density map
    shape = detector.pixel.shape
    densities_position_y, densities_position_x = trap_densities_position
    capacities_position_y, capacities_position_x = trap_capacities_position

    # Extract trap density / full well
    trap_densities_2d = load_cropped_and_aligned_image(
        shape=shape,
        filename=trap_densities_filename,
        position_x=densities_position_x,
        position_y=densities_position_y,
        align=trap_densities_align,
        allow_smaller_array=False,
    )

    if trap_capacities_filename is not None:

        # Extract the max amount of trap by long soak
        trap_capacities_2d = load_cropped_and_aligned_image(
            shape=shape,
            filename=trap_capacities_filename,
            position_x=capacities_position_x,
            position_y=capacities_position_y,
            align=trap_capacities_align,
            allow_smaller_array=False,
        )

        trap_capacities_2d[np.where(trap_capacities_2d < 0)] = 0

    else:
        trap_capacities_2d = None

    trap_densities_2d[np.where(trap_densities_2d < 0)] = 0
    trap_densities_2d = np.nan_to_num(
        trap_densities_2d, nan=0.0, posinf=0.0, neginf=0.0
    )

    if not np.all((0 <= trap_densities_2d) & (trap_densities_2d <= 1)):
        raise ValueError("Trap density map values not between 0 and 1.")

    if not detector.has_persistence():
        detector.persistence = Persistence(
            trap_time_constants=trap_time_constants,
            trap_proportions=trap_proportions,
            geometry=detector.pixel.shape,
        )

    new_pixel_array, new_all_trapped_charge = compute_persistence(
        pixel_array=detector.pixel.array,
        all_trapped_charge=detector.persistence.trapped_charge_array,
        trap_proportions=np.array(trap_proportions),
        trap_time_constants=np.array(trap_time_constants),
        trap_densities_2d=trap_densities_2d,
        trap_capacities_2d=trap_capacities_2d,
        delta_t=detector.time_step,
    )

    detector.pixel.array = new_pixel_array
    detector.persistence.trapped_charge_array = new_all_trapped_charge


@numba.njit(fastmath=True)
def compute_persistence(
    pixel_array: np.ndarray,
    all_trapped_charge: np.ndarray,
    trap_proportions: np.ndarray,
    trap_time_constants: np.ndarray,
    trap_densities_2d: np.ndarray,
    delta_t: float,
    trap_capacities_2d: t.Optional[np.ndarray] = None,
) -> t.Tuple[np.ndarray, np.ndarray]:
    """Compute persistence.

    Parameters
    ----------
    pixel_array: ndarray
    all_trapped_charge: ndarray
    trap_proportions: ndarray
    trap_time_constants: ndarray
    trap_densities_2d: ndarray
    delta_t: float
    trap_capacities_2d: ndarray, optional

    Returns
    -------
    tuple of ndarray
    """
    pixel_start = pixel_array.copy()

    for i, trapped_charge in enumerate(all_trapped_charge):

        # Computer trapped charge for this increment of time
        # Time factor is the integration time divided by the time constant (1, 10, 100, 1000, 10000)
        time_factor = delta_t / trap_time_constants[i]

        # Amount of charges trapped per unit of full well
        densities = trap_densities_2d * trap_proportions[i]

        # Compute trapped charges
        available_traps = densities * pixel_array
        empty_traps = available_traps - trapped_charge
        diff = time_factor * empty_traps

        # maximum you can release is trapped_charge, max you can add is empty traps
        diff = clip_diff(
            diff=diff, trapped_charge=trapped_charge, empty_traps=empty_traps
        )

        trapped_charge += diff

        pixel_array -= diff

        all_trapped_charge[i] = trapped_charge

    pixel_diff = pixel_array - pixel_start

    for i, trapped_charge in enumerate(all_trapped_charge):

        if trap_capacities_2d is None:
            fwc = None
        else:
            fwc = trap_capacities_2d * trap_proportions[i]

        densities = trap_densities_2d * trap_proportions[i]
        available_traps = pixel_array * densities

        trapped_charge_clipped, output_pixel = clip_trapped_charge(
            trapped_charge=trapped_charge,
            pixel=pixel_array,
            available_traps=available_traps,
            pixel_diff=pixel_diff,
            trap_capacities=fwc,
        )
        all_trapped_charge[i] = trapped_charge_clipped

    return output_pixel, all_trapped_charge


@numba.njit(fastmath=True)
def clip_trapped_charge(
    trapped_charge: np.ndarray,
    pixel: np.ndarray,
    available_traps: np.ndarray,
    pixel_diff: np.ndarray,
    trap_capacities: t.Optional[np.ndarray] = None,
) -> t.Tuple[np.ndarray, np.ndarray]:
    """Clip trapped charge between arrays of max and min value.

    Parameters
    ----------
    trapped_charge: ndarray
    pixel: ndarray
    available_traps: ndarray
    pixel_diff: ndarray
    trap_capacities: ndarray, optional

    Returns
    -------
    output: tuple of ndarray
    """

    n, m = trapped_charge.shape
    clipped = trapped_charge.copy()
    pixel_output = pixel.copy()

    for i in range(n):
        for j in range(m):
            if pixel_diff[i, j] < 0:
                if trap_capacities is not None:
                    maximum = min(available_traps[i, j], trap_capacities[i, j])
                else:
                    maximum = available_traps[i, j]
                if trapped_charge[i, j] > maximum:
                    clipped[i, j] = maximum

    pixel_output += trapped_charge - clipped

    return clipped, pixel_output


@numba.njit(fastmath=True)
def clip_diff(
    diff: np.ndarray, trapped_charge: np.ndarray, empty_traps: np.ndarray
) -> np.ndarray:
    """Clip diff array between arrays of min and max values.

    Parameters
    ----------
    diff: ndarray
    trapped_charge: ndarray
    empty_traps: ndarray

    Returns
    -------
    output: ndarray
    """
    n, m = diff.shape
    output = diff.copy()
    for i in range(n):
        for j in range(m):
            if diff[i, j] < 0:
                if diff[i, j] < -trapped_charge[i, j]:
                    output[i, j] = -trapped_charge[i, j]
            else:
                if diff[i, j] > empty_traps[i, j]:
                    output[i, j] = empty_traps[i, j]
    return output



from collections import abc
from pathlib import Path

import pytest

import pyxel
from pyxel import Configuration
from pyxel.detectors import CCD
from pyxel.exposure import run_exposure_pipeline
from pyxel.observation.observation import Observation, ParameterMode
from pyxel.pipelines import DetectionPipeline, Processor

try:
    import pygmo as pg

    WITH_PYGMO = True
except ImportError:
    WITH_PYGMO = False


expected_sequential = [
    (0, [("level", 10), ("initial_energy", 100)]),
    (1, [("level", 20), ("initial_energy", 100)]),
    (2, [("level", 30), ("initial_energy", 100)]),
    (3, [("level", 100), ("initial_energy", 100)]),
    (4, [("level", 100), ("initial_energy", 200)]),
    (5, [("level", 100), ("initial_energy", 300)]),
]

expected_product = [
    (0, [("level", 10), ("initial_energy", 100)]),
    (1, [("level", 10), ("initial_energy", 200)]),
    (2, [("level", 10), ("initial_energy", 300)]),
    (3, [("level", 20), ("initial_energy", 100)]),
    (4, [("level", 20), ("initial_energy", 200)]),
    (5, [("level", 20), ("initial_energy", 300)]),
    (6, [("level", 30), ("initial_energy", 100)]),
    (7, [("level", 30), ("initial_energy", 200)]),
    (8, [("level", 30), ("initial_energy", 300)]),
]


@pytest.mark.skipif(not WITH_PYGMO, reason="Package 'pygmo' is not installed.")
@pytest.mark.parametrize(
    "mode, expected",
    [
        # ('single', expected_single),
        (ParameterMode.Sequential, expected_sequential),
        (ParameterMode.Product, expected_product),
    ],
)
def test_pipeline_parametric_without_init_photon(mode: ParameterMode, expected):
    input_filename = "tests/data/parametric.yaml"
    cfg = pyxel.load(Path(input_filename))

    assert isinstance(cfg, Configuration)
    assert hasattr(cfg, "observation")
    assert hasattr(cfg, "ccd_detector")
    assert hasattr(cfg, "pipeline")

    observation = cfg.observation
    assert isinstance(observation, Observation)

    observation.parameter_mode = mode

    detector = cfg.ccd_detector
    assert isinstance(detector, CCD)

    pipeline = cfg.pipeline
    assert isinstance(pipeline, DetectionPipeline)

    processor = Processor(detector=detector, pipeline=pipeline)
    result = observation.debug_parameters(processor)
    assert result == expected

    processor_generator = observation._processors_it(processor=processor)
    assert isinstance(processor_generator, abc.Generator)

    for proc, _, _ in processor_generator:
        assert isinstance(proc, Processor)

        run_exposure_pipeline(processor=proc, readout=observation.readout)



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.
#
#
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.


from pathlib import Path

import pytest
from freezegun import freeze_time

import pyxel


@pytest.fixture
def config_no_running_mode(tmp_path: Path) -> Path:
    """Create a configuration file without a valid running mode."""
    content = """
ccd_detector:

  geometry:

    row: 450               # pixel
    col: 450               # pixel
    total_thickness: 40.    # um
    pixel_vert_size: 10.    # um
    pixel_horz_size: 10.    # um

  environment:
    temperature: 300        # K

  characteristics:
    qe:   1.                # -
    eta:  1.                # e/photon
    sv:   3.e-6             # V/e
    amp:  10.               # V/V
    a1:   100.              # V/V
    a2:   3000              # DN/V
    fwc:  2000              # e
    fwc_serial: 10000       # e
    vg:   1.62e-10          # cm^2
    svg:  1.62e-10          # cm^2
    t:    9.4722e-04        # s
    st:   9.4722e-04        # s

pipeline:
  photon_generation:
    - name: illumination
      func: pyxel.models.photon_generation.illumination
      enabled: true
      arguments:
          level: 0
  optics:
  charge_generation:
  charge_collection:
  charge_transfer:
  charge_measurement:
  readout_electronics:
"""

    filename = tmp_path / "no_running_mode.yaml"

    with filename.open(mode="w") as fh:
        fh.write(content)

    return filename


@pytest.fixture
def config_two_running_modes(tmp_path: Path) -> Path:
    """Create a configuration file with two valid running mode."""
    content = """
exposure:

  outputs:
    output_folder: "output"
    save_data_to_file:
      - detector.image.array:   ['fits']
      - detector.pixel.array: ['npy']
      
observation:

  parameters:
    - key: pipeline.photon_generation.illumination.arguments.level
      values: numpy.unique(numpy.logspace(0, 6, 10, dtype=int))

  outputs:
    output_folder:  'outputs'
    # each pipeline run
    save_data_to_file:
      - detector.image.array:   ['npy']
    # once at the end
    save_observation_data:


ccd_detector:

  geometry:

    row: 450               # pixel
    col: 450               # pixel
    total_thickness: 40.    # um
    pixel_vert_size: 10.    # um
    pixel_horz_size: 10.    # um

  environment:
    temperature: 300        # K

  characteristics:
    qe:   1.                # -
    eta:  1.                # e/photon
    sv:   3.e-6             # V/e
    amp:  10.               # V/V
    a1:   100.              # V/V
    a2:   3000              # DN/V
    fwc:  2000              # e
    fwc_serial: 10000       # e
    vg:   1.62e-10          # cm^2
    svg:  1.62e-10          # cm^2
    t:    9.4722e-04        # s
    st:   9.4722e-04        # s

pipeline:
  photon_generation:
    - name: illumination
      func: pyxel.models.photon_generation.illumination
      enabled: true
      arguments:
          level: 0
  optics:
  charge_generation:
  charge_collection:
  charge_transfer:
  charge_measurement:
  readout_electronics:
"""

    filename = tmp_path / "no_running_mode.yaml"

    with filename.open(mode="w") as fh:
        fh.write(content)

    return filename


@pytest.fixture
def config_no_detector(tmp_path: Path) -> Path:
    """Create a configuration file without detector."""
    content = """
exposure:

  outputs:
    output_folder: "output"
    save_data_to_file:
      - detector.image.array:   ['fits']
      - detector.pixel.array: ['npy']

pipeline:
  photon_generation:
    - name: illumination
      func: pyxel.models.photon_generation.illumination
      enabled: true
      arguments:
          level: 0
  optics:
  charge_generation:
  charge_collection:
  charge_transfer:
  charge_measurement:
  readout_electronics:
"""

    filename = tmp_path / "no_running_mode.yaml"

    with filename.open(mode="w") as fh:
        fh.write(content)

    return filename


@pytest.fixture
def config_two_detectors(tmp_path: Path) -> Path:
    """Create a configuration file with two detectors."""
    content = """
ccd_detector:

  geometry:

    row: 450               # pixel
    col: 450               # pixel
    total_thickness: 40.    # um
    pixel_vert_size: 10.    # um
    pixel_horz_size: 10.    # um

  environment:
    temperature: 300        # K

  characteristics:
    qe:   1.                # -
    eta:  1.                # e/photon
    sv:   3.e-6             # V/e
    amp:  10.               # V/V
    a1:   100.              # V/V
    a2:   3000              # DN/V
    fwc:  2000              # e
    fwc_serial: 10000       # e
    vg:   1.62e-10          # cm^2
    svg:  1.62e-10          # cm^2
    t:    9.4722e-04        # s
    st:   9.4722e-04        # s

cmos_detector:

  geometry:

    row: 100               # pixel
    col: 100               # pixel
    total_thickness: 40.    # um
    pixel_vert_size: 10.    # um
    pixel_horz_size: 10.    # um
    n_output: 1
    n_row_overhead: 0
    n_frame_overhead: 0
    reverse_scan_direction: False
    reference_pixel_border_width: 4

  environment:
    temperature: 300

  characteristics:
    qe: 0.5               # -
    eta: 1                # e/photon
    sv: 1.0e-6            # V/e
    amp: 0.8              # V/V
    a1: 100               # V/V
    a2: 50000             # DN/V
    #a2: 65536             # DN/V
    fwc: 100000            # e
    dsub: 0.5

exposure:

  outputs:
    output_folder: "output"
    save_data_to_file:
      - detector.image.array:   ['fits']
      - detector.pixel.array: ['npy']

pipeline:
  photon_generation:
    - name: illumination
      func: pyxel.models.photon_generation.illumination
      enabled: true
      arguments:
          level: 0
  optics:
  charge_generation:
  charge_collection:
  charge_transfer:
  charge_measurement:
  readout_electronics:
"""

    filename = tmp_path / "no_running_mode.yaml"

    with filename.open(mode="w") as fh:
        fh.write(content)

    return filename


def test_load_2_times():
    """Test function 'pyxel.load' called two times."""
    filename = "tests/data/dummy_simple.yaml"

    # Get full filename
    full_filename = Path(filename)  # type: Path
    assert full_filename.exists()

    # Load the configuration file for the first time
    with freeze_time("2021-06-15 14:11"):
        _ = pyxel.load(full_filename)

    # Load the configuration file for the second time
    with freeze_time("2021-06-15 14:11"):
        _ = pyxel.load(full_filename)


def test_load_no_running_mode(config_no_running_mode: Path):
    """test function 'pyxel.load' without a running mode."""
    filename = config_no_running_mode

    with pytest.raises(
        ValueError,
        match=r"Expecting only one running mode: 'exposure', 'observation', 'calibration'",
    ):
        _ = pyxel.load(filename)


def test_load_two_running_modes(config_two_running_modes: Path):
    """test function 'pyxel.load' without two running modes."""
    filename = config_two_running_modes

    with pytest.raises(
        ValueError,
        match=r"Expecting only one running mode: 'exposure', 'observation', 'calibration'",
    ):
        _ = pyxel.load(filename)


def test_load_no_detector(config_no_detector: Path):
    """test function 'pyxel.load' without detector."""
    filename = config_no_detector

    with pytest.raises(
        ValueError,
        match=r"Expecting only one detector: 'ccd_detector', 'cmos_detector', 'mkid_detector'",
    ):
        _ = pyxel.load(filename)


def test_load_two_detectors(config_two_detectors: Path):
    """test function 'pyxel.load' with two detectors."""
    filename = config_two_detectors

    with pytest.raises(
        ValueError,
        match=r"Expecting only one detector: 'ccd_detector', 'cmos_detector', 'mkid_detector'",
    ):
        _ = pyxel.load(filename)



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

""":term:`SAR` :term:`ADC` model."""
import numpy as np

from pyxel.detectors import Detector


def apply_sar_adc(
    signal_2d: np.ndarray,
    num_rows: int,
    num_cols: int,
    min_volt: float,
    max_volt: float,
    adc_bits: int,
) -> np.ndarray:
    """Apply :term:`SAR` :term:`ADC`.

    Parameters
    ----------
    signal_2d : ndarray
    num_rows : int
    num_cols : int
    min_volt : float
    max_volt : float
    adc_bits : int

    Returns
    -------
    ndarray
    """
    data_digitized_2d = np.zeros((num_rows, num_cols))

    # First normalize the data to voltage since there is no model for
    # the conversion photogenerated carrier > volts in the model/charge_measure
    signal_normalized_2d = signal_2d * max_volt / np.max(signal_2d)

    # Set the reference voltage of the ADC to half the max
    ref = max_volt / 2.0  # type: float

    # For each bits, compare the value of the ref to the capacitance value
    for i in np.arange(adc_bits):
        # digital value associated with this step
        digital_value = 2 ** (adc_bits - (i + 1))

        # All data that is higher than the ref is equal to the dig. value
        data_digitized_2d[signal_normalized_2d >= ref] += digital_value

        # Subtract ref value from the data
        signal_normalized_2d[signal_normalized_2d >= ref] -= ref

        # Divide reference voltage by 2 for next step
        ref /= 2.0

    return data_digitized_2d


# TODO: documentation, range volt - only max is used
def sar_adc(
    detector: Detector,
) -> None:
    """Digitize signal array using :term:`SAR` (Successive Approximation Register) :term:`ADC` logic.

    Parameters
    ----------
    detector : Detector
        Pyxel Detector object.
    """

    min_volt, max_volt = detector.characteristics.adc_voltage_range
    adc_bits = detector.characteristics.adc_bit_resolution

    image_2d = apply_sar_adc(
        signal_2d=detector.signal.array,
        num_rows=detector.geometry.row,
        num_cols=detector.geometry.col,
        min_volt=min_volt,
        max_volt=max_volt,
        adc_bits=adc_bits,
    )  # type: np.ndarray

    detector.image.array = image_2d



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""The detector attribute and data structure subpackage."""

# flake8: noqa
# Warning: Import order matters
from .environment import Environment
from .detector import Detector
from .ccd import CCD
from .cmos import CMOS
from .mkid import MKID
from .geometry import Geometry
from .ccd_geometry import CCDGeometry
from .cmos_geometry import CMOSGeometry
from .mkid_geometry import MKIDGeometry
from .characteristics import Characteristics
from .ccd_characteristics import CCDCharacteristics
from .cmos_characteristics import CMOSCharacteristics
from .mkid_characteristics import MKIDCharacteristics
from .optics import Optics
from .readout_properties import ReadoutProperties



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""TBW."""
import functools
import inspect
import typing as t

if t.TYPE_CHECKING:
    from pyxel.detectors import Detector


# Declare type variable
T = t.TypeVar("T")


class Arguments(t.MutableMapping):
    """Arguments class for usage in ModelFunction.

    Class Arguments is initialized from a dictionary of model function arguments.
    It resembles the behaviour of the passed dictionary with additional methods __setattr__ and __getattr__,
    which enable to get and set the model parameters through attribute interface.
    Dictionary of arguments is saved in private attribute _arguments.

    Examples
    --------
    >>> from pyxel.pipelines.model_function import Arguments
    >>> arguments = Arguments({"one": 1, "two": 2})
    >>> arguments
    Arguments({'one': 1, 'two': 2})

    Access arguments
    >>> arguments["one"]
    1
    or
    >>> arguments.one
    1

    Changing parameters
    >>> arguments["one"] = 10
    >>> arguments["two"] = 20
    >>> arguments
    Arguments({'one': 10, 'two': 20})

    Non existing arguments
    >>> arguments["three"] = 3
    KeyError: 'No argument named three !'
    >>> arguments.three = 3
    AttributeError: 'No argument named three !'
    """

    def __init__(self, input_arguments: dict):
        self._arguments = dict(input_arguments)

    def __setitem__(self, key, value):

        if key not in self._arguments:
            raise KeyError(f"No argument named {key} !")

        self._arguments[key] = value

    def __getitem__(self, key):

        if key not in self._arguments:
            raise KeyError(f"No argument named {key} !")

        return self._arguments[key]

    def __delitem__(self, key):
        del self._arguments[key]

    def __iter__(self):
        return iter(self._arguments)

    def __len__(self):
        return len(self._arguments)

    def __getattr__(self, key):

        # Use non-modified __getattr__ in this case.
        if key == "_arguments":
            return object.__getattribute__(self, "_arguments")

        if key not in self._arguments:
            raise AttributeError(f"'{self!r}' object has no argument {key!r} !")

        return self._arguments[key]

    def __setattr__(self, key, value):

        # Use non-modified __setattr__ in this case.
        if key == "_arguments":
            super().__setattr__(key, value)
            return

        if key not in self._arguments:
            raise AttributeError(f"'{self!r}' object has no argument {key!r} !")

        self._arguments[key] = value

    def __dir__(self):
        return dir(type(self)) + list(self)

    def __repr__(self):
        return f"Arguments({self._arguments})"

    # def __deepcopy__(self, memo) -> "Arguments":
    #     """TBW."""
    #     return Arguments(deepcopy(self._arguments))


# TODO: Improve this class. See issue #132.
# TODO: Add unit tests #182
class ModelFunction:
    """Create a wrapper function around a Model function.

    Examples
    --------
    >>> from pyxel.models.photon_generation.illumination import illumination
    >>> model_func = ModelFunction(
    ...     func=illumination,
    ...     name="illumination",
    ...     arguments={"level": 1, "option": "foo"},
    ... )

    Access basic parameters

    >>> model_func.name
    'illumination'
    >>> model_func.enabled
    True
    >>> model_func.arguments
    Arguments({'level': 1, 'option': 'foo'})

    Access the arguments with a ``dict`` interface

    >>> list(model_func.arguments)
    ['level', 'option']
    >>> model_func.arguments["level"]
    1
    >>> model_func.arguments["level"] = 2
    TypeError: 'Arguments' object does not support item assignment

    Access the arguments with an attribute interface

    >>> model_func.arguments.level
    1
    """

    def __init__(
        self,
        func: t.Callable,
        name: str,
        arguments: t.Optional[dict] = None,
        enabled: bool = True,
    ):
        if inspect.isclass(func):
            raise AttributeError("Cannot pass a class to ModelFunction.")

        self._func = func  # type: t.Callable
        self._name = name
        self.enabled = enabled  # type: bool

        if arguments is None:
            arguments = {}

        self._arguments = Arguments(arguments)
        # self.group = None               # TODO

    def __repr__(self) -> str:
        cls_name = self.__class__.__name__  # type: str
        func_name = self._func.__module__ + "." + self._func.__name__

        return (
            f"{cls_name}(name={self.name!r}, func={func_name}, "
            f"arguments={self.arguments!r}, enabled={self.enabled!r})"
        )

    @property
    def name(self) -> str:
        """TBW."""
        return self._name

    @property
    def arguments(self) -> Arguments:
        """TBW."""
        return self._arguments

    def __call__(self, detector: "Detector") -> T:
        """TBW."""
        # func_ref = evaluate_reference(self.func)  # type: t.Callable

        # if inspect.isclass(self._func):
        #     # this is a class type, instantiate it using default arguments.
        #     func_ref = self._func()
        #     # TODO: should check whether or not it's callable.
        #     raise NotImplementedError
        #
        # else:
        #     func_ref = self._func

        func = functools.partial(self._func, **self.arguments)

        return func(detector)



"""Unittests for class `Environment`."""
#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

import pytest

from pyxel.detectors import Environment


@pytest.mark.parametrize(
    "list_in, list_out",
    [
        # (None, None, None, None, None, None),
        # (1, None, None, 1.0, None, None),
        # (None, 2, None, None, 2.0, None),
        # (None, None, 3, None, None, 3.0),
        ([1], [1.0]),
        ([4.0], [4.0]),
        ([7], [7.0]),
    ],
)
def test_init(list_in, list_out):
    """Test Environment.__init__ with valid entries."""

    # Create the object
    obj = Environment(
        temperature=list_in[0],
    )

    # Test getters
    assert obj.temperature == list_out[0]


# @pytest.mark.parametrize("temperature, total_inonising_dose, total_non_ionising_dose, exp_exception, exp_msg", [
#     ('foo', 2.0, 3.0, TypeError, "Unsupported type for 'temperature"),
#     (1.0 * u.K, 2.0, 3.0, TypeError, "Unsupported type for 'temperature"),
#     (1.0 * u.C, 2.0, 3.0, TypeError, "Unsupported type for 'temperature"),
#     (1.0 * u.m, 2.0, 3.0, TypeError, "Unsupported type for 'temperature"),
#
#     (1.0, 'foo', 3.0, TypeError, "Unsupported type for 'total_ionising_dose"),
#     (1.0, 2.0, 'foo', TypeError, "Unsupported type for 'total_non_ionising_dose"),
#
#     (0, 2.0, 3.0, ValueError, "'temperature' must be > 0"),
#     (0.0, 2.0, 3.0, ValueError, "'temperature' must be > 0"),
#     (-0.1, 2.0, 3.0, ValueError, "'temperature' must be > 0"),
#     (-1, 2.0, 3.0, ValueError, "'temperature' must be > 0"),
#
#     (1.0, -0.1, 3.0, ValueError, "'total_ionising_dose' must be >= 0"),
#     (1.0, -1, 3.0, ValueError, "'total_ionising_dose' must be >= 0"),
#
#     (1.0, 2.0, -0.1, ValueError, "'total_non_ionising_dose' must be >= 0"),
#     (1.0, 2.0, -1, ValueError, "'total_non_ionising_dose' must be >= 0"),
# ])
# def test_failed_init(temperature, total_inonising_dose, total_non_ionising_dose, exp_exception, exp_msg):
#     """Test Environment.__init__ with wrong inputs."""
#     with pytest.raises(exp_exception) as exc_info:
#         _ = Environment(temperature=temperature,
#                         total_ionising_dose=total_inonising_dose,
#                         total_non_ionising_dose=total_non_ionising_dose)
#     assert exc_info.match(exp_msg)
#
#
# @pytest.mark.parametrize("obj, other_obj", [
#     (Environment(), Environment()),
#     (Environment(1.0), Environment(1.0)),
#     (Environment(1.0, total_ionising_dose=2.0), Environment(temperature=1, total_ionising_dose=2)),
#     (Environment(1.0, 2.0, total_non_ionising_dose=3.0), Environment(1.0, 2.0, 3.0)),
#     (Environment(temperature=1.0, total_ionising_dose=2.0, total_non_ionising_dose=3.0), Environment(1, 2, 3)),
# ])
# def test_eq(obj, other_obj):
#     """Test Environment.__eq__."""
#     assert type(obj) is Environment
#
#     assert obj == other_obj
#
#
# @pytest.mark.parametrize("obj, other_obj", [
#     (Environment(), None),
#     (Environment(), Environment(1, 2, 3)),
#     (Environment(1, 2, 3), Environment(4, 2, 3)),
#     (Environment(1, 2, 3), Environment(1, 4, 3)),
#     (Environment(1, 2, 3), Environment(1, 2, 4)),
# ])
# def test_neq(obj, other_obj):
#     """Test Environment.__neq__."""
#     assert type(obj) is Environment
#
#     assert obj != other_obj
#
#
# @pytest.mark.parametrize("obj", [
#     Environment(),
#     Environment(temperature=1.0),
#     Environment(temperature=1.0, total_ionising_dose=2.0),
#     Environment(temperature=1.0, total_ionising_dose=2.0, total_non_ionising_dose=3.0),
# ])
# def test_copy(obj):
#     """Test Environment.copy."""
#     assert type(obj) is Environment
#     id_obj = id(obj)
#
#     new_obj = obj.copy()
#     assert id_obj == id(obj)
#     assert obj is not new_obj
#
#     assert obj == new_obj
#
#
# @pytest.mark.parametrize("obj, exp_state", [
#     (Environment(), {'temperature': None,
#                      'total_ionising_dose': None,
#                      'total_non_ionising_dose': None}),
#     (Environment(1, 2, 3), {'temperature': 1.0,
#                             'total_ionising_dose': 2.0,
#                             'total_non_ionising_dose': 3.0}),
#     (Environment(temperature=1), {'temperature': 1.0,
#                                   'total_ionising_dose': None,
#                                   'total_non_ionising_dose': None}),
#     (Environment(total_ionising_dose=2.0), {'temperature': None,
#                                             'total_ionising_dose': 2.0,
#                                             'total_non_ionising_dose': None}),
#     (Environment(total_non_ionising_dose=3.0), {'temperature': None,
#                                                 'total_ionising_dose': None,
#                                                 'total_non_ionising_dose': 3.0}),
# ])
# def test_getstate(obj, exp_state):
#     """Test Environment.__getstate__."""
#     assert type(obj) is Environment
#
#     state = obj.__getstate__()
#     assert state == exp_state
#
#
# @pytest.mark.parametrize("state, exp_obj", [
#     ({'temperature': None,
#       'total_ionising_dose': None,
#       'total_non_ionising_dose': None}, Environment()),
#     ({'temperature': 1.0,
#       'total_ionising_dose': 2.0,
#       'total_non_ionising_dose': 3.0}, Environment(1.0, 2.0, 3.0)),
#     ({'temperature': 1.0,
#       'total_ionising_dose': None,
#       'total_non_ionising_dose': None}, Environment(temperature=1.0)),
#     ({'temperature': None,
#       'total_ionising_dose': 2.0,
#       'total_non_ionising_dose': None}, Environment(total_ionising_dose=2.0)),
#     ({'temperature': None,
#       'total_ionising_dose': None,
#       'total_non_ionising_dose': 3.0}, Environment(total_non_ionising_dose=3.0)),
#
# ])
# def test_setstate(state, exp_obj):
#     """Test Environment.__setstate__."""
#     assert isinstance(state, dict)
#
#     new_obj = Environment()
#     new_obj.__setstate__(state)
#
#     assert new_obj is not exp_obj
#     assert new_obj == exp_obj
#
#
# @pytest.mark.parametrize("obj, exp_repr", [
#     (Environment(), "Environment(temperature=None, total_ionising_dose=None, total_non_ionising_dose=None)"),
#     (Environment(1.0, 2.0, 3.0), "Environment(temperature=1.0,
#     total_ionising_dose=2.0, total_non_ionising_dose=3.0)"),
#
# ])
# def test_repr(obj, exp_repr):
#     """Test Environment.__repr__."""
#     assert type(obj) is Environment
#
#     assert repr(obj) == exp_repr



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Phase-pulse processing model."""

import numpy as np

from pyxel.detectors import MKID


def convert_to_phase(
    array: np.ndarray,
    wavelength: float,
    responsivity: float,
    scaling_factor: float = 2.5e2,
) -> np.ndarray:
    """Convert an array of charge into phase.

    Parameters
    ----------
    array: ndarray
    wavelength: float
    responsivity: float
    scaling_factor: float

    Returns
    -------
    ndarray
    """
    if not wavelength > 0:
        raise ValueError("Only positive values accepted for wavelength.")
    if not scaling_factor > 0:
        raise ValueError("Only positive values accepted for scaling_factor.")
    if not responsivity > 0:
        raise ValueError("Only positive values accepted for responsivity.")

    output = array * wavelength * scaling_factor / responsivity
    return output.astype("float64")


def pulse_processing(
    detector: MKID,
    wavelength: float,
    responsivity: float,
    scaling_factor: float = 2.5e2,
) -> None:
    """TBW.

    Parameters
    ----------
    detector: MKID
        Pyxel :term:`MKID` detector object.
    wavelength: float
        Wavelength.
    responsivity: float
        Responsivity of the pixel.
    scaling_factor: float
        Scaling factor taking into account the missing pieces of superconducting physics,
        as well as the resonator quality factor, the bias power,
        the quasi-particle losses, etc.
    """
    if not isinstance(detector, MKID):
        raise TypeError("Expecting a MKID object for the detector.")

    detector.phase.array = convert_to_phase(
        array=detector.charge.array,
        wavelength=wavelength,
        responsivity=responsivity,
        scaling_factor=scaling_factor,
    )



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""The pipeline code for the different detector simulation routines."""

# flake8: noqa
from .model_function import ModelFunction, Arguments
from .model_group import ModelGroup
from .pipeline import DetectionPipeline
from .processor import Processor, ResultType, result_keys



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Readout electronics model."""

from pyxel.detectors import MKID


def simple_phase_conversion(detector: MKID, phase_conversion: float = 1.0) -> None:
    """Create an image array from phase array.

    Parameters
    ----------
    detector: MKID
        Pyxel :term:`MKID` detector object.
    phase_conversion : float
        Phase conversion factor
    """
    if not isinstance(detector, MKID):
        raise TypeError("Expecting a MKID object for the detector.")

    detector.image.array = phase_conversion * detector.phase.array



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Pyxel wrapper class for NGHXRG - Teledyne HxRG Noise Generator model."""

import logging
import typing as t
from dataclasses import dataclass

import numpy as np
from typing_extensions import Literal

from pyxel.detectors import CMOS, CMOSGeometry
from pyxel.models.charge_measurement.nghxrg.nghxrg_beta import HXRGNoise
from pyxel.util import temporary_random_state


@dataclass
class KTCBiasNoise:
    """KTCBiasNoise parameters."""

    ktc_noise: float
    bias_offset: float
    bias_amp: float


@dataclass
class WhiteReadNoise:
    """WhiteReadNoise parameters."""

    rd_noise: float
    ref_pixel_noise_ratio: float


@dataclass
class CorrPinkNoise:
    """CorrPinkNoise parameters."""

    c_pink: float


@dataclass
class UncorrPinkNoise:
    """UncorrPinkNoise parameters."""

    u_pink: float


@dataclass
class AcnNoise:
    """AcnNoise parameters."""

    acn: float


@dataclass
class PCAZeroNoise:
    """PCAZeroNoise parameters."""

    pca0_amp: float


# Create an alias
NoiseType = t.Union[
    KTCBiasNoise,
    WhiteReadNoise,
    CorrPinkNoise,
    UncorrPinkNoise,
    AcnNoise,
    PCAZeroNoise,
]


# TODO: Use function `simcado.nghxrg.HXRGNoise` instead of
#       `pyxel.models.charge_measurement.nghxrg.nghxrg.HXRGNoise`
def compute_nghxrg(
    pixel_2d: np.ndarray,
    noise: t.Sequence[NoiseType],
    detector_shape: t.Tuple[int, int],
    window_pos: t.Tuple[int, int],
    window_size: t.Tuple[int, int],
    num_outputs: int,
    time_step: int,
    num_rows_overhead: int,
    num_frames_overhead: int,
    reverse_scan_direction: bool,
    reference_pixel_border_width: int,
) -> np.ndarray:
    """Compute NGHXRG.

    Parameters
    ----------
    pixel_2d : ndarray
    noise : sequence of NoiseType
    detector_shape : int, int
    window_pos : int, int
    window_size : int, int
    num_outputs : int
    time_step : int
    num_rows_overhead : int
    num_frames_overhead : int
    reverse_scan_direction : bool
    reference_pixel_border_width : int

    Returns
    -------
    ndarray
    """
    det_size_y, det_size_x = detector_shape
    window_y_start, window_x_start = window_pos
    window_y_size, window_x_size = window_size

    if window_pos == (0, 0) and window_size == detector_shape:
        window_mode = "FULL"  # type: Literal["FULL", "WINDOW"]
    else:
        window_mode = "WINDOW"

    data_hxrg_2d = np.asarray(pixel_2d.copy(), dtype=float)  # type: np.ndarray

    ng = HXRGNoise(
        n_out=num_outputs,
        time_step=time_step,
        nroh=num_rows_overhead,
        nfoh=num_frames_overhead,
        reverse_scan_direction=reverse_scan_direction,
        reference_pixel_border_width=reference_pixel_border_width,
        pca0=data_hxrg_2d,
        det_size_x=det_size_x,
        det_size_y=det_size_y,
        wind_mode=window_mode,
        wind_x_size=window_x_size,
        wind_y_size=window_y_size,
        wind_x0=window_x_start,
        wind_y0=window_y_start,
        verbose=True,
    )  # type: HXRGNoise

    final_data_2d = np.zeros(shape=window_size)

    for item in noise:  # type: NoiseType
        if isinstance(item, KTCBiasNoise):
            data = ng.add_ktc_bias_noise(
                ktc_noise=item.ktc_noise,
                bias_offset=item.bias_offset,
                bias_amp=item.bias_amp,
            )  # type: np.ndarray

        elif isinstance(item, WhiteReadNoise):
            data = ng.add_white_read_noise(
                rd_noise=item.rd_noise,
                reference_pixel_noise_ratio=item.ref_pixel_noise_ratio,
            )

        elif isinstance(item, CorrPinkNoise):
            data = ng.add_corr_pink_noise(c_pink=item.c_pink)

        elif isinstance(item, UncorrPinkNoise):
            data = ng.add_uncorr_pink_noise(u_pink=item.u_pink)

        elif isinstance(item, AcnNoise):
            data = ng.add_acn_noise(acn=item.acn)

        elif isinstance(item, PCAZeroNoise):
            data = ng.add_pca_zero_noise(pca0_amp=item.pca0_amp)

        else:
            raise TypeError(f"Unknown item: {item!r} !")

        data_2d = ng.format_result(data)  # type: np.ndarray
        if data_2d.any():
            if window_mode == "FULL":
                final_data_2d += data_2d
            elif window_mode == "WINDOW":
                window_y_end = window_y_start + window_y_size  # type: int
                window_x_end = window_x_start + window_x_size  # type: int

                final_data_2d[
                    window_y_start:window_y_end, window_x_start:window_x_end
                ] += data_2d

    return final_data_2d


# TODO: copyright
@temporary_random_state
def nghxrg(
    detector: CMOS,
    noise: t.Sequence[
        t.Mapping[
            Literal[
                "ktc_bias_noise",
                "white_read_noise",
                "corr_pink_noise",
                "uncorr_pink_noise",
                "acn_noise",
                "pca_zero_noise",
            ],
            t.Mapping[str, float],
        ]
    ],
    window_position: t.Optional[t.Tuple[int, int]] = None,
    window_size: t.Optional[t.Tuple[int, int]] = None,
    n_output: int = 1,
    n_row_overhead: int = 0,
    n_frame_overhead: int = 0,
    reverse_scan_direction: bool = False,
    reference_pixel_border_width: int = 4,
    seed: t.Optional[int] = None,  # This parameter is used by '@temporary_random_state'
) -> None:
    """Generate fourier noise power spectrum on HXRG detector.

    For more information see :cite:p:`2015:rauscher`.

    Parameters
    ----------
    detector: Detector
    noise: list
    window_position: t.Sequence, optional
        [x0 (columns), y0 (rows)].
    window_size: t.Sequence, optional
        [x (columns), y (rows)].
    seed: int, optional
    n_output: int
        Number of detector outputs.
    n_row_overhead: int
        New row overhead in pixel.
        This allows for a short wait at the end of a row before starting the next row.
    n_frame_overhead: int
        New frame overhead in rows.
        This allows for a short wait at the end of a frame before starting the next frame.
    reverse_scan_direction: bool
        Set this True to reverse the fast scanner readout directions.
        This capability was added to support Teledyne’s programmable fast scan readout directions.
        The default setting (False) corresponds to what HxRG detectors default to upon power up.
    reference_pixel_border_width: int
        Width of reference pixel border around image area.
    """
    if n_output not in range(33):
        raise ValueError("'n_output' must be between 0 and 32.")

    if n_row_overhead not in range(101):
        raise ValueError("'n_row_overhead' must be between 0 and 100.")

    if n_frame_overhead not in range(101):
        raise ValueError("'n_frame_overhead' must be between 0 and 100.")

    if reference_pixel_border_width not in range(33):
        raise ValueError("'reference_pixel_border_width' must be between 0 and 32.")

    # Converter
    params = []  # type: t.List[NoiseType]
    for item in noise:
        if "ktc_bias_noise" in item:
            sub_item = item["ktc_bias_noise"]  # type: t.Mapping[str, float]
            param = KTCBiasNoise(
                ktc_noise=sub_item["ktc_noise"],
                bias_offset=sub_item["bias_offset"],
                bias_amp=sub_item["bias_amp"],
            )  # type: NoiseType

        elif "white_read_noise" in item:
            sub_item = item["white_read_noise"]
            param = WhiteReadNoise(
                rd_noise=sub_item["rd_noise"],
                ref_pixel_noise_ratio=sub_item["ref_pixel_noise_ratio"],
            )

        elif "corr_pink_noise" in item:
            sub_item = item["corr_pink_noise"]
            param = CorrPinkNoise(c_pink=sub_item["c_pink"])

        elif "uncorr_pink_noise" in item:
            sub_item = item["uncorr_pink_noise"]
            param = UncorrPinkNoise(u_pink=sub_item["u_pink"])

        elif "acn_noise" in item:
            sub_item = item["acn_noise"]
            param = AcnNoise(acn=sub_item["acn"])

        elif "pca_zero_noise" in item:
            sub_item = item["pca_zero_noise"]
            param = PCAZeroNoise(pca0_amp=sub_item["pca0_amp"])

        else:
            raise KeyError(f"Unknown key: item = {item!r}")

        params.append(param)

    logging.getLogger("nghxrg").setLevel(logging.WARNING)

    # Prepare the parameters
    geo = detector.geometry  # type: CMOSGeometry

    if window_position is None:
        window_position = (0, 0)
    if window_size is None:
        window_size = (geo.row, geo.col)

    if detector.is_dynamic:
        time_step = int(detector.time / detector.time_step)  # type: int
    else:
        time_step = 1

    # Compute new pixels
    result_2d = compute_nghxrg(
        pixel_2d=detector.pixel.array,
        noise=params,
        detector_shape=(geo.row, geo.col),
        window_pos=window_position,
        window_size=window_size,
        num_outputs=n_output,
        time_step=time_step,
        num_rows_overhead=n_row_overhead,
        num_frames_overhead=n_frame_overhead,
        reverse_scan_direction=reverse_scan_direction,
        reference_pixel_border_width=reference_pixel_border_width,
    )  # type: np.ndarray

    # Add the pixels
    detector.pixel.array += result_2d


# TODO: This generates plot. It should be in class `Output`
# def display_noisepsd(
#     array: np.ndarray,
#     nb_output: int,
#     dimensions: t.Tuple[int, int],
#     noise_name: str,
#     path: Path,
#     step: int,
#     # mode: Literal["plot"] = "plot",
# ) -> t.Tuple[t.Any, np.ndarray]:
#     """Display noise PSD from the generated FITS file.
#
#     For power spectra, need to be careful of readout directions.
#
#     For the periodogram, using Welch's method to smooth the PSD
#     > Divide data in N segment of length nperseg which overlaps at nperseg/2
#     (noverlap argument)
#     nperseg high means less averaging for the PSD but more points.
#     """
#     # Conversion gain
#     conversion_gain = 1
#     data_corr = array * conversion_gain
#
#     if nb_output <= 1:
#         raise ValueError("Parameter 'nb_output' must be >= 1.")
#
#     # Should be in the simulated data header
#     dimension = dimensions[0]  # type: int
#     pix_p_output = dimension ** 2 / nb_output  # Number of pixels per output
#     nbcols_p_channel = dimension / nb_output  # Number of columns per channel
#     nperseg = int(pix_p_output / 10.0)  # Length of segments for Welch's method
#     read_freq = 100000  # Frame rate [Hz]
#
#     # Initializing table of nb_outputs periodogram +1 for dimensions
#     pxx_outputs = np.zeros((nb_output, int(nperseg / 2) + 1))
#
#     # For each output
#     for i in np.arange(nb_output):
#         # If i odd, flatten data since its the good reading direction
#         if i % 2 == 0:
#             start_x_idx = int(i * nbcols_p_channel)
#             end_x_idx = int((i + 1) * nbcols_p_channel)
#
#             output_data = data_corr[:, start_x_idx:end_x_idx].flatten()
#         # Else, flip it left/right and then flatten it
#         else:
#             start_x_idx = int(i * nbcols_p_channel)
#             end_x_idx = int((i + 1) * nbcols_p_channel)
#
#             output_data = np.fliplr(data_corr[:, start_x_idx:end_x_idx])
#             output_data.flatten()
#         # output data without flipping
#         # output_data = data_corr[:,int(i*(nbcols_p_channel)):
#         #                         int((i+1)*(nbcols_p_channel))].flatten()
#
#         # print(output_data, read_freq, nperseg)
#         # Add periodogram to the previously initialized array
#         f_vect, pxx_outputs[i] = signal.welch(output_data, read_freq, nperseg=nperseg)
#
#     # For the detector
#     # detector_data = data_corr.flatten()
#     # Add periodogram to the previously initialized array
#     # test, Pxx_detector = signal.welch(detector_data,
#     #                                   read_freq,
#     #                                   nperseg=nperseg)
#
#     # if mode == "plot":
#     #     fig, ax1 = plt.subplots(1, 1, figsize=(8, 8))
#     #     fig.canvas.set_window_title("Power Spectral Density")
#     #     fig.suptitle(
#     #         noise_name
#     #         + " Power Spectral Density\n"
#     #         + "Welch seg. length / Nb pixel output: "
#     #         + str("{:1.2f}".format(nperseg / pix_p_output))
#     #     )
#     #     ax1.plot(f_vect, np.mean(pxx_outputs, axis=0), ".-", ms=3, alpha=0.5, zorder=32)
#     #     for _, ax in enumerate([ax1]):
#     #         ax.set_xlim([1, 1e5])
#     #         ax.set_xscale("log")
#     #         ax.set_yscale("log")
#     #         ax.set_xlabel("Frequency [Hz]")
#     #         ax.set_ylabel("PSD [e-${}^2$/Hz]")
#     #         ax.grid(True, alpha=0.4)
#     #
#     #     filename = path.joinpath(
#     #         "nghxrg_" + noise_name + "_" + str(step) + ".png"
#     #     )  # type: Path
#     #     plt.savefig(filename, dpi=300)
#     #     plt.close()
#
#     result = np.asarray(np.mean(pxx_outputs, axis=0))  # type: np.ndarray
#     return f_vect, result



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Script to run Pyxel as a module.

Examples
--------
$ python -m pyxel --version
Pyxel, version ...

$ python -m pyxel --help
...

$ python -m pyxel --config my_config.yaml
...
"""

from pyxel.run import main

if __name__ == "__main__":
    main(prog_name="python -m pyxel")



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Pyxel util functions for Particle classes."""

import math
import typing as t

import numpy as np


def check_energy(initial_energy: t.Union[int, float]) -> None:
    """Check energy of the particle if it is a float or int.

    :param initial_energy:
    :return:
    """
    if not isinstance(initial_energy, (int, float)):
        raise TypeError("Given particle energy could not be read")


def random_direction(
    v_abs: float = 1.0, seed: t.Optional[int] = None
) -> np.ndarray:  # TODO check random angles and direction
    """Generate random direction for a photon.

    Parameters
    ----------
    v_abs: float
    seed: int
    """
    rng = np.random.default_rng(seed=seed)
    alpha = 2 * math.pi * rng.random()
    beta = 2.0 * math.pi * rng.random()
    v_z = v_abs * math.sin(alpha)
    v_ver = v_abs * math.cos(alpha) * math.cos(beta)
    v_hor = v_abs * math.cos(alpha) * math.sin(beta)
    return np.array([v_ver, v_hor, v_z])



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.
#
#
"""Util functions to handle random seeds."""

import typing as t
from contextlib import contextmanager
from functools import wraps

import numpy as np


@contextmanager
def change_random_state(seed: int):
    """TBW.

    Parameters
    ----------
    seed: int
    """
    state = np.random.get_state()
    np.random.seed(seed)
    try:
        yield
    finally:
        np.random.set_state(state)


def temporary_random_state(func: t.Callable) -> t.Callable:
    """Temporarily change numpy random seed within a function.

    Parameters
    ----------
    func: callable

    Returns
    -------
    inner: callable
    """

    @wraps(func)
    def inner(*args, seed=None, **kwargs):
        if seed is None:
            return func(*args, seed=seed, **kwargs)
        else:
            with change_random_state(seed):
                return func(*args, seed=seed, **kwargs)

    return inner



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""TBW."""
import typing as t

import numpy as np

from pyxel.util.memory import get_size


class Characteristics:
    """Characteristic attributes of the detector.

    Parameters
    ----------
    quantum_efficiency: float, optional
        Quantum efficiency.
    charge_to_volt_conversion: float, optional
        Sensitivity of charge readout. Unit: V/e-
    pre_amplification: float, optional
        Gain of pre-amplifier. Unit: V/V
    full_well_capacity: float, optional
        Full well capacity. Unit: e-
    adc_voltage_range: tuple of floats, optional
        ADC voltage range. Unit: V
    adc_bit_resolution: int, optional
        ADC bit resolution.
    """

    def __init__(
        self,
        quantum_efficiency: t.Optional[float] = None,  # unit: NA
        charge_to_volt_conversion: t.Optional[float] = None,  # unit: volt/electron
        pre_amplification: t.Optional[float] = None,  # unit: V/V
        full_well_capacity: t.Optional[float] = None,  # unit: electron
        adc_bit_resolution: t.Optional[int] = None,
        adc_voltage_range: t.Optional[t.Tuple[float, float]] = None,  # unit: V
    ):
        if quantum_efficiency and not (0.0 <= quantum_efficiency <= 1.0):
            raise ValueError("'quantum_efficiency' must be between 0.0 and 1.0.")
        if charge_to_volt_conversion and not (
            0.0 <= charge_to_volt_conversion <= 100.0
        ):
            raise ValueError(
                "'charge_to_volt_conversion' must be between 0.0 and 100.0."
            )
        if pre_amplification and not (0.0 <= pre_amplification <= 10000.0):
            raise ValueError("'pre_amplification' must be between 0.0 and 10000.0.")
        if full_well_capacity and not (0.0 <= full_well_capacity <= 1.0e7):
            raise ValueError("'full_well_capacity' must be between 0 and 1e7.")
        if adc_bit_resolution and not (4 <= adc_bit_resolution <= 64):
            raise ValueError("'adc_bit_resolution' must be between 4 and 64.")
        if adc_voltage_range and not len(adc_voltage_range) == 2:
            raise ValueError("Voltage range must have length of 2.")

        self._quantum_efficiency = quantum_efficiency
        self._charge_to_volt_conversion = charge_to_volt_conversion
        self._pre_amplification = pre_amplification
        self._full_well_capacity = full_well_capacity
        self._adc_voltage_range = adc_voltage_range
        self._adc_bit_resolution = adc_bit_resolution

        self._numbytes = 0

    @property
    def quantum_efficiency(self) -> float:
        """Get Quantum efficiency."""
        if self._quantum_efficiency:
            return self._quantum_efficiency
        else:
            raise ValueError(
                "'quantum_efficiency' not specified in detector characteristics."
            )

    @quantum_efficiency.setter
    def quantum_efficiency(self, value: float) -> None:
        """Set Quantum efficiency."""
        if np.min(value) < 0.0 and np.max(value) <= 1.0:
            raise ValueError("'quantum_efficiency' values must be between 0.0 and 1.0.")

        self._quantum_efficiency = value

    @property
    def charge_to_volt_conversion(self) -> float:
        """Get charge to volt conversion parameter."""
        if self._charge_to_volt_conversion:
            return self._charge_to_volt_conversion
        else:
            raise ValueError(
                "'charge_to_volt_conversion' not specified in detector characteristics."
            )

    @charge_to_volt_conversion.setter
    def charge_to_volt_conversion(self, value: float) -> None:
        """Set charge to volt conversion parameter."""
        if not (0.0 <= value <= 100.0):
            raise ValueError(
                "'charge_to_volt_conversion' must be between 0.0 and 100.0."
            )
        self._charge_to_volt_conversion = value

    @property
    def pre_amplification(self) -> float:
        """Get voltage pre-amplification gain."""
        if self._pre_amplification:
            return self._pre_amplification
        else:
            raise ValueError(
                "'pre_amplification' not specified in detector characteristics."
            )

    @pre_amplification.setter
    def pre_amplification(self, value: float) -> None:
        """Set voltage pre-amplification gain.."""
        if not (0.0 <= value <= 10000.0):
            raise ValueError("'pre_amplification' must be between 0.0 and 10000.0.")

        self._pre_amplification = value

    @property
    def adc_bit_resolution(self) -> int:
        """Get bit resolution of the Analog-Digital Converter."""
        if self._adc_bit_resolution:
            return self._adc_bit_resolution
        else:
            raise ValueError(
                "'adc_bit_resolution' not specified in detector characteristics."
            )

    @adc_bit_resolution.setter
    def adc_bit_resolution(self, value: int) -> None:
        """Set bit resolution of the Analog-Digital Converter."""
        self._adc_bit_resolution = value

    @property
    def adc_voltage_range(self) -> t.Tuple[float, float]:
        """Get voltage range of the Analog-Digital Converter."""
        if self._adc_voltage_range:
            return self._adc_voltage_range
        else:
            raise ValueError(
                "'adc_voltage_range' not specified in detector characteristics."
            )

    @adc_voltage_range.setter
    def adc_voltage_range(self, value: t.Tuple[float, float]) -> None:
        """Set voltage range of the Analog-Digital Converter."""
        self._adc_voltage_range = value

    @property
    def full_well_capacity(self) -> float:
        """Get Full well capacity."""
        if self._full_well_capacity:
            return self._full_well_capacity
        else:
            raise ValueError(
                "'full_well_capacity' not specified in detector characteristics."
            )

    @full_well_capacity.setter
    def full_well_capacity(self, value: float) -> None:
        """Set Full well capacity."""
        if value not in range(10000001):
            raise ValueError("'full_well_capacity' must be between 0 and 1e+7.")

        self._full_well_capacity = value

    @property
    def numbytes(self) -> int:
        """Recursively calculates object size in bytes using Pympler library.

        Returns
        -------
        int
            Size of the object in bytes.
        """
        self._numbytes = get_size(self)
        return self._numbytes



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""TBW."""

# flake8: noqa
from .amplification import simple_amplifier
from .sar_adc import sar_adc
from .amplifier_crosstalk import dc_crosstalk, ac_crosstalk
from .dead_time import dead_time_filter
from .phase_conversion import simple_phase_conversion
from .simple_adc import simple_adc



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Pyxel full well models."""
import typing as t

import numpy as np

from pyxel.detectors import Detector


def apply_simple_full_well_capacity(array: np.ndarray, fwc: int) -> np.ndarray:
    """Apply full well capacity to an array.

    Parameters
    ----------
    array: ndarray
    fwc: int

    Returns
    -------
    output: ndarray
    """
    array[array > fwc] = fwc
    return array


def simple_full_well(detector: Detector, fwc: t.Optional[int] = None) -> None:
    """Limit the amount of charge in pixel due to full well capacity.

    Uses full well capacity in the characteristics of the detector object if not overridden by the function argument.

    Parameters
    ----------
    detector: Detector
    fwc: int
    """
    if fwc is None:
        fwc_input = detector.characteristics.full_well_capacity
    else:
        fwc_input = fwc

    if fwc_input < 0:
        raise ValueError("Full well capacity should be a positive number.")

    charge_array = apply_simple_full_well_capacity(
        array=detector.pixel.array, fwc=fwc_input
    )

    detector.pixel.array = charge_array



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Pyxel Photon class to generate and track photon."""

import typing as t
import warnings

import numpy as np

from pyxel.data_structure import Array

if t.TYPE_CHECKING:
    from pyxel.detectors import Geometry


class Photon(Array):
    """Photon class defining and storing information of all photon.

    Accepted array types: ``np.int32``, ``np.int64``, ``np.uint32``, ``np.uint64``,
    ``np.float16``, ``np.float32``, ``np.float64``
    """

    # TODO: add unit (ph)
    EXP_TYPE = float
    TYPE_LIST = (
        np.dtype(np.float16),
        np.dtype(np.float32),
        np.dtype(np.float64),
    )

    def __init__(self, geo: "Geometry"):
        new_array = np.zeros((geo.row, geo.col), dtype=self.EXP_TYPE)

        super().__init__(new_array)

    @property
    def array(self) -> np.ndarray:
        """Two dimensional numpy array storing the data.

        Only accepts an array with the right type and shape.
        """
        return super().array

    @array.setter
    def array(self, value: np.ndarray) -> None:
        """Overwrite the two dimensional numpy array storing the data.

        Only accepts an array with the right type and shape.
        """
        self.validate_type(value)
        self.validate_shape(value)

        if np.any(value < 0):
            value[value < 0] = 0.0
            warnings.warn(
                "Trying to set negative values in the Photon array! Negative values clipped to 0."
            )

        self._array = value
